#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import warnings
from covsirphy import ChangeFinder, Trend, Term


class TestChangeFinder(object):
    @pytest.mark.parametrize(
        "country",
        [
            "Italy", "Japan", "India", "United States", "Greece", "Russia",
            "Brazil", "France", "Spain", "UK", "New Zealand", "Germany",
        ]
    )
    def test_find(self, jhu_data, population_data, country, max_rmsle=20.0):
        # Setup
        population = population_data.value(country)
        sr_df = jhu_data.to_sr(country=country, population=population)
        # Find change points
        change_finder = ChangeFinder(sr_df, max_rmsle=max_rmsle)
        change_finder.run()
        # For all phases, check if RMSLE score is lower than max_rmsle=20.0
        for (start_date, end_date) in zip(*change_finder.date_range()):
            phase_df = jhu_data.to_sr(
                country=country, population=population,
                start_date=start_date, end_date=end_date)
            rmsle = Trend(sr_df=phase_df).rmsle()
            assert rmsle < max_rmsle

    @pytest.mark.parametrize("country", ["Italy"])
    def test_show(self, jhu_data, population_data, country):
        warnings.simplefilter("ignore", category=DeprecationWarning)
        warnings.simplefilter("ignore", category=UserWarning)
        population = population_data.value(country)
        sr_df = jhu_data.to_sr(country=country, population=population)
        change_finder = ChangeFinder(sr_df)
        change_finder.run()
        change_finder.show(area=jhu_data.area_name(country=country))
        assert isinstance(change_finder.date_range(), tuple)

    def test_find_with_small_min_size(self, jhu_data, population_data):
        population = population_data.value("Italy")
        sr_df = jhu_data.to_sr(country="Italy", population=population)
        with pytest.raises(ValueError):
            change_finder = ChangeFinder(sr_df, min_size=2)
            change_finder.run()

    def test_find_with_few_records(self, jhu_data, population_data):
        population = population_data.value("Italy")
        min_size = 7
        df = jhu_data.subset(country="Italy")
        start_date = df.loc[df.index[0], Term.DATE].strftime(Term.DATE_FORMAT)
        end_date = Term.date_change(start_date, days=min_size - 2)
        sr_df = jhu_data.to_sr(
            country="Italy", population=population, end_date=end_date)
        with pytest.raises(ValueError):
            change_finder = ChangeFinder(sr_df, min_size=min_size)
            change_finder.run()


class TestTrend(object):
    @pytest.mark.parametrize("country", ["Italy"])
    @pytest.mark.parametrize("func", ["linear", "negative exponential"])
    def test_one_phase(self, jhu_data, population_data, country, func):
        population = population_data.value(country)
        sr_df = jhu_data.to_sr(
            country=country, population=population,
            start_date="25Mar2020", end_date="02Apr2020")
        trend = Trend(sr_df)
        assert isinstance(trend.rmsle(), float)
        trend.show(area=jhu_data.area_name(country="Italy"))

    @pytest.mark.parametrize("country", ["Italy"])
    def test_one_phase_with_few_records(self, jhu_data, population_data, country):
        warnings.simplefilter("ignore", category=UserWarning)
        population = population_data.value(country)
        sr_df = jhu_data.to_sr(
            country=country, population=population,
            start_date="25Mar2020", end_date="26Mar2020")
        with pytest.raises(ValueError):
            Trend(sr_df).run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from covsirphy.util.plotting import line_plot
from covsirphy.cleaning.term import Term
from covsirphy.cleaning.jhu_data import JHUData
from covsirphy.cleaning.population import PopulationData


class DataHandler(Term):
    """
    Data handler for scenario analysis.

    Args:
        jhu_data (covsirphy.JHUData): object of records
        population_data (covsirphy.PopulationData): PopulationData object
        country (str): country name
        province (str or None): province name
        auto_complement (bool): if True and necessary, the number of cases will be complemented
    """

    def __init__(self, jhu_data, population_data, country, province=None, auto_complement=True):
        # Population
        population_data = self.ensure_instance(
            population_data, PopulationData, name="population_data")
        self.population = population_data.value(country, province=province)
        # Records
        self.jhu_data = self.ensure_instance(
            jhu_data, JHUData, name="jhu_data")
        # Area name
        self.country = country
        self.province = province or self.UNKNOWN
        self.area = JHUData.area_name(country, province)
        # Whether complement the number of cases or not
        self._auto_complement = bool(auto_complement)
        self._complemented = False
        # Create {scenario_name: PhaseSeries} and set records
        self.record_df = pd.DataFrame()
        self._first_date = None
        self._last_date = None

    def init_records(self):
        """
        Set records.
        Only when auto-complement mode, complement records if necessary.
        """
        # Set records (complement records, if necessary)
        self.record_df, self._complemented = self.jhu_data.records(
            country=self.country, province=self.province,
            start_date=self._first_date, end_date=self._last_date,
            population=self.population,
            auto_complement=self._auto_complement
        )
        # First/last date of the records
        if self._first_date is None:
            series = self.record_df.loc[:, self.DATE]
            self._first_date = series.min().strftime(self.DATE_FORMAT)
            self._last_date = series.max().strftime(self.DATE_FORMAT)

    @property
    def first_date(self):
        """
        str: the first date of the records
        """
        return self._first_date

    @first_date.setter
    def first_date(self, date):
        self.ensure_date_order(self._first_date, date, name="date")
        self.ensure_date_order(date, self._last_date, name="date")
        self._first_date = date
        self.init_records()

    @property
    def last_date(self):
        """
        str: the last date of the records
        """
        return self._last_date

    @last_date.setter
    def last_date(self, date):
        self.ensure_date_order(self._first_date, date, name="date")
        self.ensure_date_order(date, self._last_date, name="date")
        self._last_date = date
        self.init_records()

    def complement(self, interval=2, max_ignored=100):
        """
        Complement the number of recovered cases, if necessary.

        Args:
            interval (int): expected update interval of the number of recovered cases [days]
            max_ignored (int): Max number of recovered cases to be ignored [cases]

        Returns:
            covsirphy.Scenario: self
        """
        self.record_df, self._complemented = self.jhu_data.records(
            country=self.country, province=self.province,
            start_date=self._first_date, end_date=self._last_date,
            population=self.population,
            auto_complement=True, interval=interval, max_ignored=max_ignored
        )
        return self

    def complement_reverse(self):
        """
        Restore the raw records. Reverse method of covsirphy.Scenario.complement().

        Returns:
            covsirphy.Scenario: self
        """
        self.record_df, self._complemented = self.jhu_data.records(
            country=self.country, province=self.province,
            start_date=self._first_date, end_date=self._last_date,
            population=self.population,
            auto_complement=False
        )
        return self

    def records(self, variables=None, show_figure=True, filename=None, **kwargs):
        """
        Return the records as a dataframe.

        Args:
            show_figure (bool): if True, show the records as a line-plot.
            variables (list[str] or None): variables to include, Infected/Fatal/Recovered when None
            filename (str): filename of the figure, or None (show figure)
            kwargs: the other keyword arguments of covsirphy.line_plot()

        Returns:
            pandas.DataFrame
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Columns set by @variables (int)
        Notes:
            - Records with Recovered > 0 will be selected.
            - If complement was performed by Scenario.complement() or Scenario(auto_complement=True),
            The kind of complement will be added to the title of the figure.
            - @variables can be selected from Susceptible/Confirmed/Infected/Fatal/Recovered.
        """
        variables = self.ensure_list(
            variables or [self.CI, self.F, self.R],
            candidates=[self.S, *self.VALUE_COLUMNS], name="variables")
        df = self.record_df.loc[:, [self.DATE, *variables]]
        if not show_figure:
            return df
        if self._complemented:
            title = f"{self.area}: Cases over time\nwith {self._complemented}"
        else:
            title = f"{self.area}: Cases over time"
        line_plot(
            df.set_index(self.DATE), title, y_integer=True, filename=filename, **kwargs)
        return df

    def records_diff(self, variables=None, window=7, show_figure=True, filename=None, **kwargs):
        """
        Return the number of daily new cases (the first discreate difference of records).

        Args:
            variables (str or None): variables to show
            window (int): window of moving average, >= 1
            show_figure (bool): if True, show the records as a line-plot.
            filename (str): filename of the figure, or None (show figure)
            kwargs: the other keyword arguments of covsirphy.line_plot()

        Returns:
            pandas.DataFrame
                Index:
                    - Date (pd.TimeStamp): Observation date
                Columns:
                    - Confirmed (int): daily new cases of Confirmed, if calculated
                    - Infected (int):  daily new cases of Infected, if calculated
                    - Fatal (int):  daily new cases of Fatal, if calculated
                    - Recovered (int):  daily new cases of Recovered, if calculated

        Notes:
            @variables will be selected from Confirmed, Infected, Fatal and Recovered.
            If None was set as @variables, ["Confirmed", "Fatal", "Recovered"] will be used.
        """
        variables = self.ensure_list(
            variables or [self.C, self.F, self.R], candidates=self.VALUE_COLUMNS, name="variables")
        window = self.ensure_natural_int(window, name="window")
        df = self.record_df.set_index(self.DATE)[variables]
        df = df.diff().dropna()
        df = df.rolling(window=window).mean().dropna().astype(np.int64)
        if not show_figure:
            return df
        if self._complemented:
            title = f"{self.area}: Daily new cases\nwith {self._complemented}"
        else:
            title = f"{self.area}: Daily new cases"
        line_plot(df, title, y_integer=True, filename=filename, **kwargs)
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
import pytest
from covsirphy import PhaseSeries
from covsirphy import Term, PhaseUnit, SIR


class TestPhaseSeries(object):

    @pytest.mark.parametrize("country", ["Japan"])
    def test_start(self, jhu_data, population_data, country):
        # Setting
        population = population_data.value(country)
        series = PhaseSeries("01Apr2020", "22Apr2020", population)
        # Whether units are registered or not
        assert not series
        assert series.summary().empty

    @pytest.mark.parametrize("country", ["Japan"])
    def test_add_phase(self, jhu_data, population_data, country):
        # Setting
        population = population_data.value(country)
        series = PhaseSeries("01Apr2020", "01Aug2020", population)
        # Last phase when empty
        empty_phase = PhaseUnit("31Mar2020", "31Mar2020", population)
        assert series.unit(phase="last") == empty_phase
        # Add a phase with specified end date: 0th
        series.add(end_date="22Apr2020")
        # Add a phase with specified population value: 1st
        with pytest.raises(ValueError):
            series.add(end_date="22Apr2020")
        series.add(end_date="05May2020", population=int(population * 0.98))
        # Add a phase with specified the number of days: 2nd
        series.add(days=21)
        # Filling past phases and add a future phase: 3rd, 4th
        series.add(end_date="01Sep2020")
        # Add a future phase: 5th
        series.add(days=30)
        # Summary
        df = series.summary()
        base_cols = [Term.TENSE, Term.START, Term.END, Term.N]
        assert set(df.columns) == set(base_cols)
        assert series.to_dict()["0th"]["Type"] == Term.PAST
        first_len = len(df)
        assert set(df.loc["3rd", :].tolist()) == set(
            [Term.PAST, "27May2020", "01Aug2020", 123998518])
        assert set(df.loc["4th", :].tolist()) == set(
            [Term.FUTURE, "02Aug2020", "01Sep2020", 123998518])
        # Disable/enable a phase
        series.disable("0th")
        assert "0th" not in series.to_dict()
        assert len(series) == first_len - 1
        assert len([unit for unit in series]) == first_len
        assert len([unit for unit in series if unit]) == first_len - 1
        series.enable("0th")
        assert "0th" in series.to_dict()
        assert len(series) == first_len
        # Clear future phases: 4th and 5th will be deleted
        series.clear(include_past=False)
        assert "4th" not in series.to_dict()
        assert len(series) == first_len - 2
        assert series
        # Clear all phases
        series.clear(include_past=True)
        assert len(series) == 0
        assert not series
        # Filling past phases: 0th
        series.add()
        assert len(series) == 1
        # Last phase
        assert series.unit(phase="last")

    @pytest.mark.parametrize("country", ["Japan"])
    def test_trend(self, jhu_data, population_data, country):
        warnings.simplefilter("ignore", category=UserWarning)
        warnings.simplefilter("ignore", category=DeprecationWarning)
        # Setting
        population = population_data.value(country)
        sr_df = jhu_data.to_sr(country=country, population=population)
        series = PhaseSeries("01Apr2020", "01Aug2020", population)
        # S-R trend analysis
        series.trend(sr_df)
        series.trend_show(sr_df=sr_df, area=None, filename=None)
        # Un-registered phase
        with pytest.raises(KeyError):
            series.unit("100th")

    @pytest.mark.parametrize("country", ["Japan"])
    def test_add_phase_with_model(self, jhu_data, population_data, country):
        # Setting
        population = population_data.value(country)
        sr_df = jhu_data.to_sr(country=country, population=population)
        series = PhaseSeries("01Apr2020", "01Aug2020", population)
        series.trend(sr_df)
        # Add future phase with model and tau
        series.add(end_date="01Sep2020", model=SIR, tau=360)
        series.add(end_date="01Oct2020")
        length = len(series)
        assert series.to_dict()[Term.num2str(length - 2)][Term.ODE] == SIR.NAME
        assert series.to_dict()[Term.num2str(length - 1)][Term.TAU] == 360
        series.add(end_date="01Nov2020", rho=0.006)
        series.add(end_date="01Dec2020", sigma=0.011)
        assert series.to_dict()[Term.num2str(length + 1)][Term.RT] == 0.55
        assert series.to_dict()[Term.num2str(length + 1)]["1/beta [day]"] == 41

    @pytest.mark.parametrize("country", ["Japan"])
    def test_delete_phase(self, jhu_data, population_data, country):
        # Setting
        population = population_data.value(country)
        sr_df = jhu_data.to_sr(country=country, population=population)
        series = PhaseSeries("01Apr2020", "01Aug2020", population)
        series.trend(sr_df)
        first_len = len(series)
        # Deletion of 0th phase is the same as disabling 0th phase
        series.delete("0th")
        series.enable("0th")
        assert len(series) == first_len
        assert "5th" in series.to_dict()
        # Delete phase (not the last registered phase)
        new_second = PhaseUnit(
            series.unit("2nd").start_date,
            series.unit("3rd").end_date,
            series.unit("2nd").population)
        series.delete("3rd")
        assert len(series) == first_len - 1
        assert series.unit("2nd") == new_second
        # Delete the last phase
        old_last = series.unit("last")
        series.delete("last")
        series.add()
        assert series.unit("last").start_date == old_last.start_date

    @pytest.mark.parametrize("country", ["Japan"])
    def test_replace(self, jhu_data, population_data, country):
        # Setting
        population = population_data.value(country)
        series = PhaseSeries("01Apr2020", "01Aug2020", population)
        series.add(end_date="01May2020")
        series.add(end_date="01Jun2020")
        series.add(end_date="01Jul2020")
        series.add()
        # Replace one old phase with one new phase
        unit_old = series.unit("0th")
        unit_new = PhaseUnit(
            unit_old.start_date, unit_old.end_date, population
        )
        unit_new.set_ode(tau=360)
        series.replace("0th", unit_new)
        assert series.unit("0th") == unit_new
        # Replace one old phase with two new phases

        unit_old = series.unit("1st")
        change_date = Term.date_change(unit_old.end_date, days=-7)
        unit_pre = PhaseUnit(
            unit_old.start_date, Term.yesterday(change_date), population)
        unit_pre.set_ode(tau=360)
        unit_fol = PhaseUnit(change_date, unit_old.end_date, population)
        unit_fol.set_ode(tau=360)
        series.replaces(phase="1st", new_list=[unit_pre, unit_fol])
        print(series.unit("1st"), unit_pre)
        assert series.unit("1st") == unit_pre
        assert series.unit("2nd") == unit_fol
        # TypeError of new_list
        with pytest.raises(TypeError):
            series.replaces(phase="3rd", new_list=[unit_pre, Term])
        # ValueError with tense
        with pytest.raises(ValueError):
            future_unit = PhaseUnit("01Sep2020", "01Dec2020", population)
            series.replaces(phase="3rd", new_list=[future_unit])
        # Add phase without deletion of any phases
        new1 = PhaseUnit("02Aug2020", "01Sep2020", population)
        new2 = PhaseUnit("02Sep2020", "01Oct2020", population)
        series.replaces(phase=None, new_list=[new1, new2])
        assert series.unit("last") == new2

    @pytest.mark.parametrize("country", ["Japan"])
    def test_simulate(self, jhu_data, population_data, country):
        # Setting
        population = population_data.value(country)
        record_df = jhu_data.subset(country, population=population)
        series = PhaseSeries("01Apr2020", "01Aug2020", population)
        # Simulation
        series.add(
            end_date="22Apr2020", model=SIR, tau=360, rho=0.006, sigma=0.011)
        df = series.simulate(record_df=record_df)
        assert set(df.columns) == set(Term.NLOC_COLUMNS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime
import warnings
import pandas as pd
import pytest
from covsirphy import ScenarioNotFoundError, UnExecutedError
from covsirphy import Scenario, DataHandler
from covsirphy import Term, PhaseSeries, SIR, SIRF


class TestDataHandler(object):
    @pytest.mark.parametrize("country", ["Italy", "Japan", "Netherlands", "Greece"])
    @pytest.mark.parametrize("province", [None, "Abruzzo"])
    def test_start(self, jhu_data, population_data, country, province):
        if province == "Abruzzo" and country != "Italy":
            with pytest.raises(KeyError):
                DataHandler(
                    jhu_data, population_data, country, province=province)
            return
        DataHandler(jhu_data, population_data, country, province=province)

    @pytest.mark.parametrize("country", ["Japan"])
    def test_start_record_range(self, jhu_data, population_data, country):
        # Setting
        dhl = DataHandler(jhu_data, population_data, country)
        dhl.init_records()
        # Test
        dhl.first_date = "01Apr2020"
        assert dhl.first_date == "01Apr2020"
        dhl.last_date = "01May2020"
        assert dhl.last_date == "01May2020"
        with pytest.raises(ValueError):
            dhl.first_date = "01Jan2019"
        with pytest.raises(ValueError):
            tomorrow = Term.tomorrow(datetime.now().strftime(Term.DATE_FORMAT))
            dhl.last_date = tomorrow

    @pytest.mark.parametrize("country", ["Japan"])
    def test_records(self, jhu_data, population_data, country):
        warnings.simplefilter("ignore", category=UserWarning)
        # Setting
        dhl = DataHandler(jhu_data, population_data, country)
        dhl.init_records()
        dhl.first_date = "01Apr2020"
        dhl.last_date = "01Aug2020"
        # Test
        df = dhl.records(variables=None, show_figure=False)
        assert isinstance(df, pd.DataFrame)
        assert set(df.columns) == set([Term.DATE, Term.CI, Term.F, Term.R])
        dates = df[Term.DATE]
        assert dates.min() == Term.date_obj(dhl.first_date)
        assert dates.max() == Term.date_obj(dhl.last_date)
        df2 = dhl.records(variables=["Susceptible"], show_figure=True)
        assert set(df2.columns) == set([Term.DATE, Term.S])

    @pytest.mark.parametrize("country", ["Japan"])
    def test_records_diff(self, jhu_data, population_data, country):
        warnings.simplefilter("ignore", category=UserWarning)
        dhl = DataHandler(jhu_data, population_data, country)
        dhl.init_records()
        dhl.records_diff(window=7, show_figure=False)
        dhl.records_diff(window=100, show_figure=True)


class TestScenario(object):
    @pytest.mark.parametrize("country", ["Japan"])
    def test_class_as_dict(self, jhu_data, population_data, country):
        # Setting
        snl = Scenario(jhu_data, population_data, country)
        # Create a phase series
        population = population_data.value(country)
        series = PhaseSeries("01Apr2020", "01Aug2020", population)
        with pytest.raises(ScenarioNotFoundError):
            assert snl["New"]
        # Add scenario
        snl["New"] = series
        # Get scenario
        assert snl["New"] == series

    @pytest.mark.parametrize("country", ["Japan"])
    def test_edit_series(self, jhu_data, population_data, country):
        # Setting
        snl = Scenario(jhu_data, population_data, country)
        snl.first_date = "01Apr2020"
        snl.last_date = "01Aug2020"
        assert snl.first_date == "01Apr2020"
        assert snl.last_date == "01Aug2020"
        # Add and clear
        assert snl.summary().empty
        snl.add(end_date="05May2020")
        with pytest.raises(ValueError):
            snl.add(end_date="20Apr2020")
        snl.add(days=20)
        snl.add()
        snl.add(end_date="01Sep2020")
        assert len(snl["Main"]) == 4
        snl.clear(include_past=True)
        with pytest.raises(ScenarioNotFoundError):
            snl.clear("New", template="Un-registered")
        snl.add(end_date="01Sep2020", name="New")
        assert len(snl["New"]) == 2
        # Delete
        snl.delete(name="Main")
        assert len(snl["Main"]) == 0
        with pytest.raises(TypeError):
            snl.delete(phases="1st", name="New")
        snl.delete(phases=["1st"], name="New")
        assert len(snl["New"]) == 1
        snl.delete(name="New")
        with pytest.raises(KeyError):
            assert len(snl["New"]) == 1

    @pytest.mark.parametrize("country", ["Japan"])
    def test_add_phase_dep(self, jhu_data, population_data, country):
        # Setting
        snl = Scenario(jhu_data, population_data, country)
        snl.first_date = "01Apr2020"
        snl.last_date = "01Aug2020"
        # Test
        warnings.simplefilter("error")
        with pytest.raises(DeprecationWarning):
            snl.add_phase(end_date="01May2020")
        warnings.simplefilter("ignore")
        snl.add_phase(end_date="01May2020")

    @pytest.mark.parametrize("country", ["Japan"])
    def test_trend(self, jhu_data, population_data, country):
        # Setting
        snl = Scenario(jhu_data, population_data, country)
        snl.first_date = "01Apr2020"
        snl.last_date = "01Aug2020"
        # Deprecated
        warnings.simplefilter("error")
        with pytest.raises(DeprecationWarning):
            snl.trend(include_init_phase=False, show_figure=False)
        warnings.simplefilter("ignore")
        snl.trend(include_init_phase=False, show_figure=False)
        # S-R trend analysis
        snl.trend(show_figure=False)
        assert snl["Main"]
        with pytest.raises(ValueError):
            snl.trend(show_figure=False, n_points=3)
        # Disable/enable
        length = len(snl["Main"])
        snl.disable(phases=["0th"], name="Main")
        assert len(snl["Main"]) == length - 1
        snl.enable(phases=["0th"], name="Main")
        assert len(snl["Main"]) == length
        with pytest.raises(TypeError):
            snl.enable(phases="0th", name="Main")
        with pytest.raises(TypeError):
            snl.disable(phases="1st", name="Main")

    @pytest.mark.parametrize("country", ["Japan"])
    def test_edit(self, jhu_data, population_data, country):
        # Setting
        snl = Scenario(jhu_data, population_data, country)
        snl.first_date = "01Apr2020"
        snl.last_date = "01Aug2020"
        snl.trend(show_figure=False)
        # Combine
        length = len(snl["Main"])
        snl.combine(["1st", "2nd"])
        n_changed = int(population_data.value(country) * 0.98)
        snl.combine(["2nd", "3rd"], population=n_changed)
        assert len(snl["Main"]) == length - 2
        # Separate
        snl.separate(date="01May2020")
        assert len(snl["Main"]) == length - 1

    @pytest.mark.parametrize("country", ["Japan"])
    def test_summary(self, jhu_data, population_data, country):
        # Setting
        snl = Scenario(jhu_data, population_data, country)
        snl.first_date = "01Apr2020"
        snl.last_date = "01Aug2020"
        snl.trend(show_figure=False)
        # One scenario
        assert set(snl.summary().columns) == set(
            [Term.TENSE, Term.START, Term.END, Term.N])
        # Show two scenarios
        snl.clear(name="New")
        cols = snl.summary().reset_index().columns
        assert set([Term.SERIES, Term.PHASE]).issubset(cols)
        # Show selected scenario
        cols_sel = snl.summary(name="New").reset_index().columns
        assert not set([Term.SERIES, Term.PHASE]).issubset(cols_sel)
        # Columns to show
        show_cols = [Term.N, Term.START]
        assert set(snl.summary(columns=show_cols).columns) == set(show_cols)
        with pytest.raises(TypeError):
            snl.summary(columns=Term.N)
        with pytest.raises(KeyError):
            snl.summary(columns=[Term.N, "Temperature"])
        # To markdown
        snl.summary().to_markdown()

    @pytest.mark.parametrize("country", ["Japan"])
    def test_estimate(self, jhu_data, population_data, country):
        warnings.simplefilter("ignore", category=UserWarning)
        # Setting
        snl = Scenario(jhu_data, population_data, country)
        snl.first_date = "01Apr2020"
        snl.last_date = "01Aug2020"
        with pytest.raises(ValueError):
            snl.estimate(SIR)
        snl.trend(show_figure=False)
        with pytest.raises(AttributeError):
            snl.estimate_history(phase="last")
        # Parameter estimation
        with pytest.raises(KeyError):
            snl.estimate(SIR, phases=["30th"])
        with pytest.raises(ValueError):
            snl.estimate(model=SIR, tau=1440)
        snl.estimate(SIR, timeout=1, timeout_iteration=1)
        # Estimation history
        snl.estimate_history(phase="last")
        # Estimation accuracy
        snl.estimate_accuracy(phase="last")
        # Get a value
        snl.get(Term.RT)
        with pytest.raises(KeyError):
            snl.get("feeling")

    @pytest.mark.parametrize("country", ["Japan"])
    def test_estimate_tau(self, jhu_data, population_data, country):
        # Setting
        snl = Scenario(jhu_data, population_data, country)
        snl.trend(show_figure=False)
        with pytest.raises(ValueError):
            snl.estimate(SIRF, tau=1440, timeout=20)

    @pytest.mark.parametrize("country", ["Japan"])
    def test_simulate(self, jhu_data, population_data, country):
        warnings.simplefilter("ignore", category=UserWarning)
        warnings.simplefilter("ignore", category=DeprecationWarning)
        # Setting
        snl = Scenario(jhu_data, population_data, country)
        snl.first_date = "01Apr2020"
        snl.last_date = "01May2020"
        with pytest.raises(ValueError):
            snl.simulate()
        with pytest.raises(ValueError):
            snl.track()
        snl.trend(show_figure=False)
        # Parameter estimation
        with pytest.raises(ValueError):
            # Deprecated
            snl.param_history(["rho"])
        all_phases = snl.summary().index.tolist()
        snl.disable(all_phases[:-2])
        with pytest.raises(NameError):
            snl.simulate()
        snl.estimate(SIRF, timeout=5, timeout_iteration=5)
        # Simulation
        snl.simulate()
        # Parameter history (Deprecated)
        snl.param_history([Term.RT], divide_by_first=False)
        snl.param_history(["rho"])
        snl.param_history(["rho"], show_figure=False)
        snl.param_history(["rho"], show_box_plot=False)
        with pytest.raises(KeyError):
            snl.param_history(["feeling"])
        # Comparison of scenarios
        snl.describe()
        snl.track()
        snl.history(target="Rt")
        snl.history(target="sigma")
        snl.history(target="rho", show_figure=False)
        snl.history(target="Infected")
        with pytest.raises(KeyError):
            snl.history(target="temperature")
        # Change rate of parameters
        snl.history_rate(name="Main")
        snl.history_rate(
            name="Main", params=["theta", "kappa"], show_figure=False)
        with pytest.raises(TypeError):
            snl.history_rate(params="", name="Main")
        # Add new scenario
        snl.add(end_date="01Sep2020", name="New")
        snl.describe()

    @pytest.mark.parametrize("country", ["Japan"])
    def test_retrospective(self, jhu_data, population_data, country):
        # Setting
        snl = Scenario(jhu_data, population_data, country)
        snl.first_date = "01Apr2020"
        snl.last_date = "01Jun2020"
        snl.trend(show_figure=False)
        # Retrospective analysis
        snl.retrospective(
            "01May2020", model=SIRF, control="Main", target="Retrospective",
            timeout=1, timeout_iteration=1)

    @pytest.mark.parametrize("country", ["Italy"])
    def test_score(self, jhu_data, population_data, country):
        snl = Scenario(jhu_data, population_data, country, tau=360)
        snl.trend(show_figure=False)
        snl.estimate(SIRF, timeout=1, timeout_iteration=1)
        assert isinstance(snl.score(metrics="RMSLE"), float)
        # Selected phases
        df = snl.summary()
        all_phases = df.index.tolist()
        sel_score = snl.score(phases=all_phases[-2:])
        # Selected past days (when the begging date is a start date)
        beginning_date = df.loc[df.index[-2], Term.START]
        past_days = Term.steps(beginning_date, snl.last_date, tau=1440)
        assert snl.score(past_days=past_days) == sel_score
        # Selected past days
        snl.score(past_days=60)
        with pytest.raises(ValueError):
            snl.score(phases=["1st"], past_days=60)

    @pytest.mark.parametrize("country", ["Japan"])
    def test_fit_predict(self, jhu_data, population_data, oxcgrt_data, country):
        snl = Scenario(jhu_data, population_data, country)
        snl.trend(show_figure=False)
        with pytest.raises(UnExecutedError):
            snl.fit(oxcgrt_data)
        snl.estimate(SIRF, timeout=1, timeout_iteration=1)
        # Fitting
        with pytest.raises(UnExecutedError):
            snl.predict()
        info_dict = snl.fit(oxcgrt_data)
        assert isinstance(info_dict, dict)
        # Prediction
        snl.predict()
        assert Term.FUTURE in snl.summary()[Term.TENSE].unique()
        # Fitting & predict
        snl.fit_predict(oxcgrt_data)
        assert Term.FUTURE in snl.summary()[Term.TENSE].unique()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections import defaultdict
from datetime import datetime, timedelta
import math
import numpy as np
import pandas as pd
from covsirphy.util.error import deprecate


class Term(object):
    """
    Term definition.
    """
    # Variables of SIR-derived model
    N = "Population"
    S = "Susceptible"
    C = "Confirmed"
    CI = "Infected"
    F = "Fatal"
    R = "Recovered"
    FR = "Fatal or Recovered"
    V = "Vaccinated"
    E = "Exposed"
    W = "Waiting"
    TESTS = "Tests"
    VAC = "Vaccinations"
    PRODUCT = "Product"
    # Column names
    DATE = "Date"
    START = "Start"
    END = "End"
    T = "Elapsed"
    TS = "t"
    TAU = "tau"
    COUNTRY = "Country"
    ISO3 = "ISO3"
    PROVINCE = "Province"
    STEP_N = "step_n"
    Y0_DICT = "y0_dict"
    PARAM_DICT = "param_dict"
    AREA_COLUMNS = [COUNTRY, PROVINCE]
    STR_COLUMNS = [DATE, *AREA_COLUMNS]
    COLUMNS = [*STR_COLUMNS, C, CI, F, R]
    NLOC_COLUMNS = [DATE, C, CI, F, R]
    SUB_COLUMNS = [DATE, C, CI, F, R, S]
    VALUE_COLUMNS = [C, CI, F, R]
    FIG_COLUMNS = [CI, F, R, FR, V, E, W]
    MONO_COLUMNS = [C, F, R]
    AREA_ABBR_COLS = [ISO3, *AREA_COLUMNS]
    # Date format: 22Jan2020 etc.
    DATE_FORMAT = "%d%b%Y"
    DATE_FORMAT_DESC = "DDMmmYYYY"
    # Separator of country and province
    SEP = "/"
    # EDA
    RATE_COLUMNS = [
        "Fatal per Confirmed",
        "Recovered per Confirmed",
        "Fatal per (Fatal or Recovered)"
    ]
    # Optimization
    A = "_actual"
    P = "_predicted"
    ACTUAL = "Actual"
    # Phase name
    SUFFIX_DICT = defaultdict(lambda: "th")
    SUFFIX_DICT.update({1: "st", 2: "nd", 3: "rd"})
    # Summary of phases
    TENSE = "Type"
    PAST = "Past"
    FUTURE = "Future"
    INITIAL = "Initial"
    ODE = "ODE"
    RT = "Rt"
    RT_FULL = "Reproduction number"
    RMSLE = "RMSLE"
    TRIALS = "Trials"
    RUNTIME = "Runtime"
    EST_COLS = [RMSLE, TRIALS, RUNTIME]
    # Scenario analysis
    PHASE = "Phase"
    SERIES = "Scenario"
    MAIN = "Main"
    # Flag
    UNKNOWN = "-"

    @classmethod
    def num2str(cls, num):
        """
        Convert numbers to 1st, 2nd etc.

        Args:
            num (int): number

        Returns:
            str
        """
        num = cls.ensure_natural_int(num, include_zero=True)
        q, mod = divmod(num, 10)
        suffix = "th" if q % 10 == 1 else cls.SUFFIX_DICT[mod]
        return f"{num}{suffix}"

    @staticmethod
    def str2num(string, name="phase names"):
        """
        Convert 1st to 1 and so on.

        Args:
            string (str): like 1st, 2nd, 3rd,...
            name (str): name of the string

        Returns:
            int
        """
        try:
            return int(string[:-2])
        except ValueError:
            raise ValueError(
                f"Examples of {name} are 0th, 1st, 2nd..., but {string} was applied.")

    @staticmethod
    def negative_exp(x, a, b):
        """
        Negative exponential function f(x)=A exp(-Bx).

        Args:
            x (float): x values
            a (float): the first parameters of the function
            b (float): the second parameters of the function

        Returns:
            float
        """
        return a * np.exp(-b * x)

    @staticmethod
    def linear(x, a, b):
        """
        Linear function f(x)=A x + b.

        Args:
            x (float): x values
            a (float): the first parameter of the function
            b (float): the second parameter of the function

        Returns:
            float
        """
        return a * x + b

    @staticmethod
    def flatten(nested_list, unique=True):
        """
        Flatten the nested list.

        Args:
            nested_list (list[list[object]]): nested list
            unique (bool): if True, only unique values will remain

        Returns:
            list[object]
        """
        flattened = sum(nested_list, list())
        if unique:
            return list(set(flattened))
        return flattened

    @staticmethod
    def ensure_dataframe(target, name="df", time_index=False, columns=None):
        """
        Ensure the dataframe has the columns.

        Args:
            target (pandas.DataFrame): the dataframe to ensure
            name (str): argument name of the dataframe
            time_index (bool): if True, the dataframe must has DatetimeIndex
            columns (list[str] or None): the columns the dataframe must have
            df (pandas.DataFrame): as-is the target
        """
        df = target.copy()
        if not isinstance(df, pd.DataFrame):
            raise TypeError(
                f"@{name} must be a instance of (pandas.DataFrame).")
        if time_index and (not isinstance(df.index, pd.DatetimeIndex)):
            raise TypeError(f"Index of @{name} must be <pd.DatetimeIndex>.")
        if columns is None:
            return df
        if not set(columns).issubset(df.columns):
            cols_str = ", ".join(
                col for col in columns if col not in df.columns)
            included = ", ".join(df.columns.tolist())
            s1 = f"Expected columns were not included in {name} with {included}."
            raise KeyError(
                f"{s1} {cols_str} must be included."
            )
        return df

    @staticmethod
    def ensure_natural_int(target, name="number", include_zero=False, none_ok=False):
        """
        Ensure a natural (non-negative) number.

        Args:
            target (int or float or str or None): value to ensure
            name (str): argument name of the value
            include_zero (bool): include 0 or not
            none_ok (bool): None value can be applied or not.

        Returns:
            int: as-is the target

        Notes:
            When @target is None and @none_ok is True, None will be returned.
            If the value is a natural number and the type was float or string,
            it will be converted to an integer.
        """
        if target is None and none_ok:
            return None
        s = f"@{name} must be a natural number, but {target} was applied"
        try:
            number = int(target)
        except TypeError:
            raise TypeError(f"{s} and not converted to integer.")
        if number != target:
            raise ValueError(f"{s}. |{target} - {number}| > 0")
        min_value = 0 if include_zero else 1
        if number < min_value:
            raise ValueError(f"{s}. This value is under {min_value}")
        return number

    @classmethod
    def ensure_tau(cls, tau):
        """
        Ensure that the value can be used as tau value [min].

        Args:
            tau (int or None): value to use

        Returns:
            int or None: as-is
        """
        if tau is None:
            return None
        tau = cls.ensure_natural_int(tau, name="tau")
        if tau in set(cls.divisors(1440)):
            return tau
        raise ValueError(
            f"@tau must be a divisor of 1440 [min], but {tau} was applied."
        )

    @classmethod
    def ensure_population(cls, population):
        """
        Ensure that the population value is valid.

        Args:
            population (int or float or str): population value

        Returns:
            int: as-is
        """
        return cls.ensure_natural_int(
            population, name="population", include_zero=False, none_ok=False
        )

    @staticmethod
    def ensure_float(target, name="value"):
        """
        Ensure a float value.
        If the value is a float value and the type was string,
        it will be converted to a float.

        Args:
            target (float or str): value to ensure
            name (str): argument name of the value

        Returns:
            float: as-is the target
        """
        s = f"@{name} must be a float value, but {target} was applied"
        try:
            value = float(target)
        except TypeError:
            raise TypeError(f"{s} and not converted to float.")
        return value

    @classmethod
    def ensure_date(cls, target, name="date"):
        """
        Ensure the format of the string.

        Args:
            target (str): string to ensure
            name (str): argument name of the string

        Returns:
            str: as-is the target
        """
        try:
            cls.date_obj(target)
        except ValueError:
            raise ValueError(
                f"@{name} needs to match {cls.DATE_FORMAT_DESC}, but {target} was applied.")
        return target

    @staticmethod
    def ensure_subclass(target, parent, name="target"):
        """
        Ensure the target is a subclass of the parent class.

        Args:
            target (object): target to ensure
            parent (object): parent class
            name (str): argument name of the target

        Returns:
            int: as-is the target
        """
        s = f"@{name} must be an sub class of {type(parent)}, but {type(target)} was applied."
        if not issubclass(target, parent):
            raise TypeError(s)
        return target

    @staticmethod
    def ensure_instance(target, class_obj, name="target"):
        """
        Ensure the target is a instance of the class object.

        Args:
            target (instance): target to ensure
            parent (class): class object
            name (str): argument name of the target

        Returns:
            object: as-is target
        """
        s = f"@{name} must be an instance of {type(class_obj)}, but {type(target)} was applied."
        if not isinstance(target, class_obj):
            raise TypeError(s)
        return target

    @staticmethod
    def ensure_list(target, candidates=None, name="target"):
        """
        Ensure the target is a sub-list of the candidates.

        Args:
            target (list[object]): target to ensure
            candidates (list[object] or None): list of candidates, if we have
            name (str): argument name of the target

        Returns:
            object: as-is target
        """
        if not isinstance(target, (list, tuple)):
            raise TypeError(
                f"@{name} must be a list or tuple, but {type(target)} was applied.")
        if candidates is None:
            return target
        # Check the target is a sub-list of candidates
        try:
            candidate_str = ", ".join(candidates)
        except TypeError:
            raise TypeError(
                f"@candidates must be a list, but {candidates} was applied.") from None
        ok_list = [element in candidates for element in target]
        if all(ok_list):
            return target
        raise KeyError(
            f"@{name} must be a sub-list of {candidate_str}, but {target} was applied.") from None

    @classmethod
    def divisors(cls, value):
        """
        Return the list of divisors of the value.

        Args:
            value (int): target value

        Returns:
            list[int]: the list of divisors
        """
        value = cls.ensure_natural_int(value)
        return [
            i for i in range(1, value + 1) if value % i == 0
        ]

    @classmethod
    def date_obj(cls, date_str=None, default=None):
        """
        Convert a string to a datatime object.

        Args:
            date_str (str or None, optional): string, like 22Jan2020
            default (datetime.datetime or None, optional): default value to return

        Returns:
            datetime.datetime or None

        Notes:
            If @date_str is None, returns @default value
        """
        if date_str is None:
            return default
        return datetime.strptime(date_str, cls.DATE_FORMAT)

    @classmethod
    def date_change(cls, date_str, days=0):
        """
        Return @days days ago or @days days later.

        Args:
            date_str (str): today
            days (int): (negative) days ago or (positive) days later

        Returns:
            str: the date
        """
        if not isinstance(days, int):
            raise TypeError(
                f"@days must be integer, but {type(days)} was applied.")
        date = cls.date_obj(date_str) + timedelta(days=days)
        return date.strftime(cls.DATE_FORMAT)

    @classmethod
    def tomorrow(cls, date_str):
        """
        Tomorrow of the date.

        Args:
            date_str (str): today

        Returns:
            str: tomorrow
        """
        return cls.date_change(date_str, days=1)

    @classmethod
    def yesterday(cls, date_str):
        """
        Yesterday of the date.

        Args:
            date_str (str): today

        Returns:
            str: yesterday
        """
        return cls.date_change(date_str, days=-1)

    @classmethod
    def steps(cls, start_date, end_date, tau):
        """
        Return the number of days (round up).

        Args:
            start_date (str): start date, like 01Jan2020
            end_date (str): end date, like 01Jan2020
            tau (int): tau value [min]
        """
        sta = cls.date_obj(start_date)
        end = cls.date_obj(end_date)
        tau = cls.ensure_tau(tau)
        return math.ceil((end - sta) / timedelta(minutes=tau))

    @classmethod
    def ensure_date_order(cls, previous_date, following_date, name="following_date"):
        """
        Ensure that the order of dates.

        Args:
            previous_date (str): previous date
            following_date (str): following date
            name (str): name of @following_date

        Raises:
            ValueError: @previous_date > @following_date
        """
        previous = cls.date_obj(previous_date)
        following = cls.date_obj(following_date)
        if previous <= following:
            return None
        raise ValueError(
            f"@{name} must be the same as/over {previous_date}, but {following_date} was applied."
        )


class Word(Term):
    @ deprecate(old="Word()", new="Term()")
    def __init__(self):
        super().__init__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from covsirphy.cleaning.cbase import CleaningBase


class OxCGRTData(CleaningBase):
    """
    Data cleaning of OxCGRT dataset.

    Args:
        filename (str): CSV filename of the dataset
        citation (str): citation
    """
    OXCGRT_VARIABLES_RAW = [
        "school_closing",
        "workplace_closing",
        "cancel_events",
        "gatherings_restrictions",
        "transport_closing",
        "stay_home_restrictions",
        "internal_movement_restrictions",
        "international_movement_restrictions",
        "information_campaigns",
        "testing_policy",
        "contact_tracing",
        "stringency_index"
    ]
    OXCGRT_COL_DICT = {v: v.capitalize() for v in OXCGRT_VARIABLES_RAW}
    OXCGRT_VARS = list(OXCGRT_COL_DICT.values())
    OXCGRT_COLS = [
        CleaningBase.DATE, CleaningBase.COUNTRY, CleaningBase.ISO3,
        *list(OXCGRT_COL_DICT.values())
    ]
    OXCGRT_COLS_WITHOUT_COUNTRY = [
        CleaningBase.DATE, *list(OXCGRT_COL_DICT.values())
    ]
    OXCGRT_VARS_INDICATORS = [
        v for v in OXCGRT_VARS if v != "Stringency_index"]

    def __init__(self, filename, citation=None):
        super().__init__(filename, citation)

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.
        This method overwrite super()._cleaning() method.
        Policy indices (Overall etc.) are from
        README.md and documentation/index_methodology.md in
        https://github.com/OxCGRT/covid-policy-tracker/

        Returns:
            pandas.DataFrame
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Country (pandas.Category): country/region name
                    - ISO3 (str): ISO 3166-1 alpha-3, like JPN
                    - other column names are defined by OxCGRTData.COL_DICT
        """
        df = self._raw.copy()
        # Rename the columns
        df = df.rename(self.OXCGRT_COL_DICT, axis=1)
        df = df.rename(
            {
                "CountryName": self.COUNTRY, "CountryCode": self.ISO3,
                "Country/Region": self.COUNTRY, "ObservationDate": self.DATE,
            },
            axis=1
        )
        df[self.COUNTRY] = df[self.COUNTRY].replace(
            {
                # COD
                "Congo, the Democratic Republic of the": "Democratic Republic of the Congo",
                # COG
                "Congo": "Republic of the Congo",
                # Diamond princess
                "Diamond Princess": "Others",
            }
        )
        # Confirm the expected columns are in raw data
        self.ensure_dataframe(
            df, name="the raw data", columns=self.OXCGRT_COLS
        )
        # Read date records
        try:
            df[self.DATE] = pd.to_datetime(df[self.DATE], format="%Y%m%d")
        except ValueError:
            df[self.DATE] = pd.to_datetime(df[self.DATE], format="%Y-%m-%d")
        # Confirm float type
        float_cols = list(self.OXCGRT_COL_DICT.values())
        for col in float_cols:
            df[col] = pd.to_numeric(df[col], errors="coerce")
            df[col] = df[col].fillna(method="ffill")
        # Select the columns to use
        df = df.loc[:, [self.DATE, self.COUNTRY, self.ISO3, *float_cols]]
        # Update data types to reduce memory
        cat_cols = [self.ISO3, self.COUNTRY]
        df[cat_cols] = df[cat_cols].astype("category")
        return df

    def subset(self, country, **kwargs):
        """
        Create a subset for a country.

        Args:
            country (str): country name or ISO 3166-1 alpha-3, like JPN
            kwargs: the other arguments will be ignored in the latest version.

        Returns:
            pandas.DataFrame
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - other column names are defined by OxCGRTData.COL_DICT
        """
        country = self.ensure_country_name(country)
        df = super().subset(country=country)
        df = df.groupby(self.DATE).last().reset_index()
        return df.loc[:, self.OXCGRT_COLS_WITHOUT_COUNTRY]

    def total(self):
        """
        This method is not defined for OxCGRTData class.
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import pandas as pd
import swifter
from covsirphy.util.error import SubsetNotFoundError
from covsirphy.cleaning.cbase import CleaningBase


class LinelistData(CleaningBase):
    """
    Linelist of case reports.

    Args:
        filename (str or pathlib.path): CSV filename to save the raw dataset
        force (bool): if True, always download the dataset from the server
        verbose (int): level of verbosity
    """
    GITHUB_URL = "https://raw.githubusercontent.com"
    URL = f"{GITHUB_URL}/beoutbreakprepared/nCoV2019/master/latest_data/latestdata.tar.gz"
    # Column names
    AGE = "Age"
    SEX = "Sex"
    HOSPITAL_DATE = "Hospitalized_date"
    CONFIRM_DATE = "Confirmation_date"
    SYMPTOM = "Symptoms"
    CHRONIC = "Chronic_disease"
    OUTCOME = "Outcome"
    OUTCOME_DATE = "Outcome_date"
    R_DATE = "Recovered_date"
    F_DATE = "Fatal_date"
    LINELIST_COLS = [
        CleaningBase.COUNTRY, CleaningBase.PROVINCE,
        HOSPITAL_DATE, CONFIRM_DATE, OUTCOME_DATE,
        CleaningBase.C, CleaningBase.CI, CleaningBase.R, CleaningBase.F,
        SYMPTOM, CHRONIC, AGE, SEX,
    ]
    # Raw dataset
    RAW_COL_DICT = {
        "age": AGE,
        "sex": SEX,
        "province": CleaningBase.PROVINCE,
        "country": CleaningBase.COUNTRY,
        "date_admission_hospital": HOSPITAL_DATE,
        "date_confirmation": CONFIRM_DATE,
        "symptoms": SYMPTOM,
        "chronic_disease": CHRONIC,
        "outcome": OUTCOME,
        "date_death_or_discharge": OUTCOME_DATE,
    }

    def __init__(self, filename, force=False, verbose=1):
        self._filename = filename
        raw_df = self._read_raw(filename, force, verbose)
        self._cleaned_df = self._cleaning(raw_df)
        self._citation = "Xu, B., Gutierrez, B., Mekaru, S. et al. " \
            "Epidemiological data from the COVID-19 outbreak, real-time case information. " \
            "Sci Data 7, 106 (2020). https://doi.org/10.1038/s41597-020-0448-0"
        # To avoid "imported but unused"
        self.__swifter = swifter

    @property
    def raw(self):
        """
        pandas.DataFrame: raw dataset
        """
        return self._read_raw(self._filename, force=False, verbose=0)

    def _read_raw(self, filename, force, verbose):
        """
        Get raw dataset from a CSV file or retrieve the dataset from server.

        Args:
            filename (str or pathlib.path): CSV filename to save the raw dataset
            force (bool): whether force retrieving from server or not when we have CSV file
            verbose (int): level of verbosity

        Returns:
            pd.DataFrame: raw dataset
        """
        if Path(filename).exists() and not force:
            return self.load(filename, columns=list(self.RAW_COL_DICT))
        Path(filename).parent.mkdir(exist_ok=True, parents=True)
        return self._retrieve(filename=filename, verbose=verbose)

    def _retrieve(self, filename, verbose=1):
        """
        Retrieve the dataset from server.

        Args:
            filename (str or pathlib.path): CSV filename to save the raw dataset
            verbose (int): level of verbosity

        Returns:
            pd.DataFrame: raw dataset
        """
        # Show URL
        if verbose:
            print(
                "Retrieving linelist from Open COVID-19 Data Working Group repository: https://github.com/beoutbreakprepared/nCoV2019")
        # Download the dataset
        df = self.load(self.URL, header=1, columns=list(self.RAW_COL_DICT))
        # Save the raw data
        df.to_csv(filename, index=False)
        return df

    def _cleaning(self, raw_df):
        """
        Perform data cleaning of the raw data.

        Args:
            raw_df (pandas.DataFrame): raw data

        Returns:
            pandas.DataFrame: cleaned data
        """
        df = raw_df.copy()
        # Rename columns
        df = df.rename(self.RAW_COL_DICT, axis=1)
        # Location
        df = df.dropna(subset=[self.COUNTRY])
        df[self.PROVINCE] = df[self.PROVINCE].fillna(self.UNKNOWN)
        # Date
        for col in [self.HOSPITAL_DATE, self.CONFIRM_DATE, self.OUTCOME_DATE]:
            df[col] = pd.to_datetime(
                df[col], infer_datetime_format=True, errors="coerce")
        df = df.loc[~df[self.CONFIRM_DATE].isna()]
        # Outcome
        df[self.C] = ~df[self.CONFIRM_DATE].isna()
        df[self.R] = df[self.OUTCOME].str.lower().isin(
            [
                "recovered", "discharge", "discharged", "released from quarantine",
                "discharged from hospital", "not hospitalized"
            ]
        )
        df[self.F] = df[self.OUTCOME].str.lower().isin(
            ["deceased", "died", "death", "dead"]
        )
        df[self.CI] = df[[self.C, self.R, self.F]].swifter.progress_bar(False).apply(
            lambda x: x[0] and not x[1] and x[2], axis=1)
        # Symptoms
        df[self.SYMPTOM] = df[self.SYMPTOM].str.replace(", ", ":")
        # Chronic disease
        df[self.CHRONIC] = df[self.CHRONIC].str.replace(", ", ":")
        df.loc[df[self.CHRONIC].str.contains(
            "http", na=False), self.CHRONIC] = None
        # Age
        df[self.AGE] = pd.to_numeric(df[self.AGE], errors="coerce")
        # Sex
        df[self.SEX] = df[self.SEX].fillna(self.UNKNOWN)
        # Update data types to reduce memory
        cat_cols = [
            self.AGE, self.SEX, self.SYMPTOM, self.CHRONIC, *self.AREA_COLUMNS]
        df[cat_cols] = df[cat_cols].astype("category")
        # Select columns
        return df.loc[:, self.LINELIST_COLS]

    def subset(self, country, province=None):
        """
        Return subset of the country/province.

        Args:
            country (str): country name or ISO3 code
            province (str or None): province name

        Returns:
            pandas.DataFrame
                Index:
                    reset index
                Columns:
                    - Hospitalized_date (pandas.TimeStamp or NT)
                    - Confirmation_date (pandas.TimeStamp or NT)
                    - Outcome_date (pandas.TimeStamp or NT)
                    - Confirmed (bool)
                    - Infected (bool)
                    - Recovered (bool)
                    - Fatal (bool)
                    - Symtoms (str)
                    - Chronic_disease (str)
                    - Age (int or None)
                    - Sex (str)
        """
        df = self._cleaned_df.copy()
        # Subset by country name
        country = self.ensure_country_name(country)
        df = df.loc[df[self.COUNTRY] == country]
        # Subset by province name
        if province not in (None, self.UNKNOWN):
            df = df.loc[df[self.PROVINCE] == province]
        # Check records are registered
        country_alias = self.ensure_country_name(country)
        if df.empty:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province)
        df = df.drop([self.COUNTRY, self.PROVINCE], axis=1)
        return df.reset_index(drop=True)

    def closed(self, outcome="Recovered"):
        """
        Return subset of global outcome data (recovered/fatal).

        Args:
            outcome (str): 'Recovered' or 'Fatal'

        Returns:
            pandas.DataFrame
                Index:
                    reset index
                Columns:
                    - Country (pandas.Category): country name
                    - Province (pandas.Category): province name or "-"
                    - Hospitalized_date (pandas.TimeStamp or NT)
                    - Confirmation_date (pandas.TimeStamp)
                    - Recovered_date (pandas.TimeStamp): if outcome is Recovered
                    - Fatal_date (pandas.TimeStamp): if outcome is Fatal
                    - Symptoms (str)
                    - Chronic_disease (str)
                    - Age (int or None)
                    - Sex (str)
        """
        column_dict = {self.R: self.R_DATE, self.F: self.F_DATE}
        if outcome not in column_dict:
            raise KeyError(
                f"@outcome should be selected from {self.R} or {self.F}, but {outcome} was applied.")
        df = self._cleaned_df.copy()
        df = df.loc[df[outcome]]
        # Subset by date
        df = df.loc[df[self.CONFIRM_DATE] < df[self.OUTCOME_DATE]]
        # Column names
        df = df.rename({self.OUTCOME_DATE: column_dict[outcome]}, axis=1)
        df = df.drop([self.C, self.CI, self.F, self.R], axis=1)
        return df.reset_index(drop=True)

    def recovery_period(self):
        """
        Calculate median value of recovery period (from confirmation to recovery).

        Returns:
            int: recovery period [days]
        """
        df = self.closed(outcome=self.R)
        series = (df[self.R_DATE] - df[self.CONFIRM_DATE]).dt.days
        return int(series.median())

    def total(self):
        """
        This is not defined for this child class.
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
import pandas as pd
import pytest
from covsirphy import PolicyMeasures
from covsirphy import SIRF, Scenario


class TestPolicyMeasures(object):
    def test_start(self, jhu_data, population_data, oxcgrt_data):
        warnings.simplefilter("ignore", category=UserWarning)
        # Create instance
        analyser = PolicyMeasures(
            jhu_data, population_data, oxcgrt_data, tau=360)
        # List of countries
        assert isinstance(analyser.countries, list)
        # Return Scenario class
        assert isinstance(analyser.scenario("Japan"), Scenario)
        with pytest.raises(KeyError):
            analyser.scenario("Moon")
        assert isinstance(analyser.countries, list)

    # Skip this test at this time
    # def test_analysis(self, jhu_data, population_data, oxcgrt_data):
    def analysis(self, jhu_data, population_data, oxcgrt_data):
        warnings.simplefilter("ignore", category=UserWarning)
        warnings.simplefilter("error", category=RuntimeWarning)
        # Create instance
        analyser = PolicyMeasures(
            jhu_data, population_data, oxcgrt_data, tau=360)
        # S-R trend analysis
        analyser.trend()
        # Phase length
        phase_len_dict = analyser.phase_len()
        assert isinstance(phase_len_dict, dict)
        assert isinstance(
            max(phase_len_dict.items(), key=lambda x: x[0])[1], list)
        # Select two countries
        phase_len_dict = analyser.phase_len()
        countries_all = [
            country
            for (_, countries) in sorted(phase_len_dict.items())
            for country in countries
        ]
        analyser.countries = countries_all[:2]
        # Parameter estimation
        with pytest.raises(ValueError):
            analyser.track()
        analyser.estimate(SIRF, timeout=1, timeout_iteration=1)
        assert isinstance(analyser.summary(), pd.DataFrame)
        # Parameter history of Rt
        with pytest.raises(KeyError):
            df = analyser.history("Temperature", roll_window=None)
        df = analyser.history("Rt", roll_window=None)
        assert isinstance(df, pd.DataFrame)
        # Parameter history of rho
        df = analyser.history("rho", roll_window=14, show_figure=False)
        assert isinstance(df, pd.DataFrame)
        # Summarize
        assert isinstance(analyser.summary(), pd.DataFrame)
        with pytest.raises(TypeError):
            analyser.summary(countries="Poland")

    def test_error(self, jhu_data, population_data, oxcgrt_data):
        warnings.simplefilter("ignore", category=UserWarning)
        warnings.simplefilter("error", category=RuntimeWarning)
        # Create instance
        analyser = PolicyMeasures(
            jhu_data, population_data, oxcgrt_data, tau=360)
        # Register countries
        with pytest.raises(KeyError):
            analyser.countries = ["Moon"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import numpy as np
import pandas as pd
from dask import dataframe as dd
import swifter
from covsirphy.util.plotting import line_plot
from covsirphy.util.error import PCRIncorrectPreconditionError, SubsetNotFoundError
from covsirphy.cleaning.cbase import CleaningBase
from covsirphy.cleaning.country_data import CountryData


class PCRData(CleaningBase):
    """
    Data cleaning of PCR dataset.

    Args:
        filename (str or None): CSV filename of the dataset
        interval (int): expected update interval of the number of confirmed cases and tests [days]
        min_pcr_tests (int): minimum number of valid daily tests performed in order to calculate positive rate
        citation (str): citation
    """
    # Column names
    PCR_VALUE_COLUMNS = [CleaningBase.TESTS, CleaningBase.C]
    PCR_NLOC_COLUMNS = [CleaningBase.DATE, *PCR_VALUE_COLUMNS]
    PCR_COLUMNS = [*CleaningBase.STR_COLUMNS, *PCR_VALUE_COLUMNS]
    # Daily values
    T_DIFF = "Tests_diff"
    C_DIFF = "Confirmed_diff"
    PCR_RATE = "Test_positive_rate"

    def __init__(self, filename, interval=2, min_pcr_tests=100, citation=None):
        if filename is None:
            self._raw = pd.DataFrame()
            self._cleaned_df = pd.DataFrame(columns=self.PCR_COLUMNS)
        else:
            self._raw = dd.read_csv(
                filename, dtype={"Province/State": "object"}
            ).compute()
            self._cleaned_df = self._cleaning()
        self.interval = self.ensure_natural_int(interval, name="interval")
        self.min_pcr_tests = self.ensure_natural_int(
            min_pcr_tests, name="min_pcr_tests")
        self._citation = citation or ""
        # Cleaned dataset of "Our World In Data"
        self._cleaned_df_owid = pd.DataFrame()
        # To avoid "imported but unused"
        self.__swifter = swifter

    def cleaned(self):
        """
        Return the cleaned dataset of PCRData with tests and confirmed data.

        Returns:
            pandas.DataFrame
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Country (pandas.Category): country/region name
                    - Province (pandas.Category): province/prefecture/state name
                    - Tests (int): the number of total tests performed
                    - Confirmed (int): the number of confirmed cases

        Notes:
            Cleaning method is defined by self._cleaning() method.
        """
        return self._cleaned_df.loc[:, self.PCR_COLUMNS]

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.
        This method overwrites super()._cleaning() method.

        Returns:
            pandas.DataFrame
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - ISO3 (str): ISO3 code
                    - Country (pandas.Category): country/region name
                    - Province (pandas.Category): province/prefecture/state name
                    - Tests (int): the number of total tests performed
                    - Confirmed (int): the number of confirmed cases
        """
        df = super()._cleaning()
        # Rename the columns
        df = df.rename(
            {
                "ObservationDate": self.DATE,
                "Country/Region": self.COUNTRY,
                "Province/State": self.PROVINCE,
            },
            axis=1
        )
        # Confirm the expected columns are in raw data
        self.ensure_dataframe(
            df, name="the raw data", columns=self.PCR_COLUMNS)
        # Datetime columns
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        # Country
        df[self.COUNTRY] = df[self.COUNTRY].replace(
            {
                # COD
                "Congo, the Democratic Republic of the": "Democratic Republic of the Congo",
                # COG
                "Congo": "Republic of the Congo",
                # South Korea
                "Korea, South": "South Korea",
            }
        )
        # Province
        df[self.PROVINCE] = df[self.PROVINCE].fillna(self.UNKNOWN)
        df.loc[df[self.COUNTRY] == "Diamond Princess", [
            self.COUNTRY, self.PROVINCE]] = ["Others", "Diamond Princess"]
        # Values
        df = df.fillna(method="ffill").fillna(0)
        df[self.TESTS] = df[self.TESTS].astype(np.int64)
        df[self.C] = df[self.C].astype(np.int64)
        df = df.loc[:, [self.ISO3, *self.PCR_COLUMNS]].reset_index(drop=True)
        # Update data types to reduce memory
        df[self.AREA_ABBR_COLS] = df[self.AREA_ABBR_COLS].astype("category")
        return df

    @classmethod
    def from_dataframe(cls, dataframe):
        """
        Create PCRData instance using a pandas dataframe.

        Args:
            dataframe (pd.DataFrame): cleaned dataset

        Returns:
            covsirphy.PCRData: PCR dataset
        """
        instance = cls(filename=None)
        instance._cleaned_df = cls.ensure_dataframe(
            dataframe, name="dataframe", columns=cls.PCR_COLUMNS)
        return instance

    def _download_ourworldindata(self, filename):
        """
        Download the dataset (ISO code/date/the number of tests) from "Our World In Data" site.
        https://github.com/owid/covid-19-data/tree/master/public/data
        https://ourworldindata.org/coronavirus

        Args:
            filename (str): CSV filename to save the datasetretrieved from "Our World In Data"
        """
        url = "https://covid.ourworldindata.org/data/testing/covid-testing-all-observations.csv"
        col_dict = {
            "ISO code": self.ISO3,
            "Date": self.DATE,
            "Cumulative total": self.TESTS,
            "Daily change in cumulative total": self.T_DIFF,
        }
        # Download the dataset
        df = self.load(url, columns=list(col_dict))
        # Data cleaning
        df = df.rename(col_dict, axis=1)
        df[self.TESTS] = pd.to_numeric(df[self.TESTS], errors="coerce")
        df[self.TESTS] = df[self.TESTS].fillna(method="ffill").astype(np.int64)
        # Calculate cumulative values if necessary
        df[self.T_DIFF] = df[self.T_DIFF].fillna(0).astype(np.int64)
        na_last_df = df.loc[
            (df[self.TESTS].isna()) & (df[self.DATE] == df[self.DATE].max())]
        re_countries_set = set(na_last_df[self.ISO3].unique())
        df["cumsum"] = df.groupby(self.ISO3)[self.T_DIFF].cumsum()
        df[self.TESTS] = df[[self.ISO3, self.TESTS, "cumsum"]].swifter.progress_bar(False).apply(
            lambda x: x[1] if x[0] in re_countries_set else x[2], axis=1)
        df = df.drop("cumsum", axis=1)
        # Drop duplicated records
        df = df.drop_duplicates(subset=[self.ISO3, self.DATE])
        # Save as CSV file
        df.to_csv(filename, index=False)
        return df

    def use_ourworldindata(self, filename, force=False):
        """
        Set the cleaned dataset retrieved from "Our World In Data" site.
        https://github.com/owid/covid-19-data/tree/master/public/data
        https://ourworldindata.org/coronavirus

        Args:
            filename (str): CSV filename to save the datasetretrieved from "Our World In Data"
            force (bool): if True, always download the dataset from "Our World In Data"
        """
        # Retrieve dataset from "Our World In Data" if necessary
        Path(filename).parent.mkdir(exist_ok=True, parents=True)
        if Path(filename).exists() and not force:
            df = self.load(filename, dtype={self.TESTS: np.int64})
        else:
            df = self._download_ourworldindata(filename)
        # Add "Country" and "Confirmed" column using "COVID-19 Data Hub" dataset
        df[self.COUNTRY] = None
        df[self.C] = None
        df.index = df[self.ISO3].str.cat(df[self.DATE], sep="_")
        series = df.loc[:, self.TESTS]
        hub_df = self._cleaned_df.copy()
        hub_df = hub_df.loc[hub_df[self.PROVINCE] == self.UNKNOWN]
        hub_df.index = hub_df[self.ISO3].str.cat(
            hub_df[self.DATE].astype(str), sep="_")
        df.update(hub_df)
        df[self.TESTS] = series
        df = df.dropna().reset_index(drop=True)
        # Add "Province" column (Unknown because not)
        df[self.PROVINCE] = self.UNKNOWN
        # Data types
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        df[self.COUNTRY] = df[self.COUNTRY].astype("category")
        df[self.PROVINCE] = df[self.PROVINCE].astype("category")
        df[self.C] = df[self.C].astype(np.int64)
        # Save the dataframe as the cleaned dataset
        self._cleaned_df_owid = df.reset_index(drop=True)
        # Update citation
        self._citation += "\nHasell, J., Mathieu, E., Beltekian, D. et al." \
            " A cross-country database of COVID-19 testing. Sci Data 7, 345 (2020)." \
            " https://doi.org/10.1038/s41597-020-00688-8"
        return self

    def replace(self, country_data):
        """
        Replace a part of cleaned dataset with a dataframe.

        Args:
            country_data (covsirphy.CountryData): dataset object of the country
                Index: reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Province (pandas.Category): province name
                    - Tests (int): the number of total tests performed
                    - Confirmed (int): the number of confirmed cases
                    - The other columns will be ignored

        Returns:
            covsirphy.PCRData: self
        """
        self.ensure_instance(country_data, CountryData, name="country_data")
        # Read new dataset
        country = country_data.country
        new = self.ensure_dataframe(
            country_data.cleaned(), name="the raw data", columns=self.PCR_COLUMNS)
        new = new.loc[:, self.PCR_COLUMNS]
        new[self.ISO3] = self.country_to_iso3(country)
        # Remove the data in the country from the current datset
        df = self._cleaned_df.copy()
        df = df.loc[df[self.COUNTRY] != country]
        # Add the new data
        df = pd.concat([df, new], axis=0, sort=False)
        # Update data types to reduce memory
        df[self.AREA_ABBR_COLS] = df[self.AREA_ABBR_COLS].astype("category")
        self._cleaned_df = df.copy()
        # Citation
        self._citation += f"\n{country_data.citation}"
        return self

    @staticmethod
    def _pcr_monotonic(df, variable):
        """
        Force the variable show monotonic increasing.

        Args:
            df (pandas.DataFrame):
                Index: Date (pandas.TimeStamp)
                Columns: Tests, Confirmed
            variable (str): variable name to show monotonic increasing

        Returns:
            pandas.DataFrame: complemented records
                Index: Date (pandas.TimeStamp)
                Columns: Tests, Confirmed
        """
        # Whether complement is necessary or not
        if df[variable].is_monotonic_increasing:
            return df
        # Complement
        decreased_dates = df[df[variable].diff() < 0].index.tolist()
        for date in decreased_dates:
            # Raw value on the decreased date
            raw_last = df.loc[date, variable]
            # Extrapolated value on the date
            series = df.loc[:date, variable]
            series.iloc[-1] = None
            series.interpolate(method="spline", order=1, inplace=True)
            series.fillna(method="ffill", inplace=True)
            # Reduce values to the previous date
            df.loc[:date, variable] = series * raw_last / series.iloc[-1]
            df[variable] = df[variable].fillna(0).astype(np.int64)
        return df

    def _pcr_check_complement(self, df, variable):
        """
        If variable values do not change for more than applied 'self.interval' days,
        indicate compliment action is needed.

        Args:
            df (pandas.DataFrame):
                Index: Date (pandas.TimeStamp)
                Columns: Tests, Confirmed, Tests_diff, C_diff
            variable: the desired column to use

        Returns:
            bool: whether complement is necessary or not
        """
        max_frequency = df[variable].value_counts().max()
        return max_frequency > self.interval or not df.loc[df.index[-1], variable]

    def _pcr_partial_complement(self, before_df, variable):
        """
        If there are missing values in variable column,
        apply partial compliment (bfill, ffill) to all columns.

        Args:
            before_df (pandas.DataFrame):
                Index: Date (pandas.TimeStamp)
                Columns: Tests, Confirmed, Tests_diff, C_diff
            variable: the desired column to use

        Returns:
            pandas.DataFrame: complemented records
                Index: Date (pandas.TimeStamp)
                Columns: Tests, Confirmed, Tests_diff, C_diff

        Notes:
            Filling NA with 0 will be always applied.
        """
        df = before_df.copy()
        df[self.TESTS].fillna(0, inplace=True)
        if self.T_DIFF in df.columns:
            df[self.T_DIFF].fillna(0, inplace=True)
        if not self._pcr_check_complement(df, variable):
            return df
        for col in df:
            df[col].replace(0, np.nan, inplace=True)
            df[col].fillna(method="ffill", inplace=True)
            df[col].fillna(method="bfill", inplace=True)
        return df

    def _pcr_processing(self, before_df, window):
        """
        Return the processed pcr data

        Args:
            before_df (pandas.DataFrame):
                Index: reset index
                Columns: Date (pandas.TimeStamp), Tests, Confirmed
            window (int): window of moving average, >= 1

        Returns:
            tuple (pandas.DataFrame, bool):
                pandas.DataFrame
                    Index: reset index
                    Columns:
                        - Date (pd.TimeStamp): Observation date
                        - Tests (int): the number of total tests performed
                        - Confirmed (int): the number of confirmed cases
                        - Tests_diff (int): daily tests performed
                        - Confirmed_diff (int): daily confirmed cases
                bool: True if complement is needed or False
        """
        df = before_df.copy()
        df[self.TESTS].fillna(method="ffill", inplace=True)
        if self.T_DIFF in df.columns:
            df[self.T_DIFF].fillna(method="ffill", inplace=True)
        # Confirmed must show monotonic increasing
        df = self._pcr_monotonic(df, self.C)
        df = self._pcr_partial_complement(df, self.TESTS)
        # If Tests values are all valid, with no missing values in-between,
        # they must be monotonically increasing as well
        df = self._pcr_monotonic(df, self.TESTS)
        # Calculate daily values for tests and confirmed (with window=1)
        df[self.T_DIFF] = df[self.TESTS].diff()
        df[self.C_DIFF] = df[self.C].diff()
        # Ensure that tests > confirmed in daily basis
        df.loc[
            df[self.T_DIFF].abs() < df[self.C_DIFF].abs(), self.T_DIFF] = None
        # Keep valid non-zero values by ignoring zeros at the beginning
        df = df.replace(0, np.nan)
        non_zero_index_start = df[self.T_DIFF].first_valid_index()
        df = df.loc[non_zero_index_start:].reset_index(drop=True)
        non_zero_index_end = df[self.T_DIFF].last_valid_index()
        # Keep valid non-zero values by complementing zeros at the end
        if non_zero_index_end < (len(df) - 1):
            df.loc[non_zero_index_end + 1:, self.T_DIFF] = None
        df = self._pcr_partial_complement(df, self.T_DIFF)
        # Use rolling window for averaging tests and confirmed
        df[self.T_DIFF] = df[self.T_DIFF].rolling(window).mean()
        df[self.C_DIFF] = df[self.C_DIFF].rolling(window).mean()
        df = self._pcr_partial_complement(df, self.T_DIFF)
        # Remove first zero lines due to window
        df = df.replace(0, np.nan)
        non_zero_index_start = df[self.T_DIFF].first_valid_index()
        df = df.loc[non_zero_index_start:].reset_index(drop=True)
        # Complemented or not
        is_complemented = df.drop(
            [self.T_DIFF, self.C_DIFF], axis=1).equals(before_df)
        return (df, is_complemented)

    def _pcr_check_preconditions(self, df):
        """
        Check preconditions in order to proceed with PCR data processing.

        Args:
            df (pandas.DataFrame):
                Index: Date (pandas.TimeStamp)
                Columns: Tests, Confirmed

        Return:
            bool: whether the dataset has sufficient data or not
        """
        df[self.TESTS].fillna(0, inplace=True)
        if self.T_DIFF in df.columns:
            df[self.T_DIFF].fillna(0, inplace=True)
        # Check if the values are zero or nan
        check_zero = df[self.TESTS].max()
        # Check if the number of the missing values
        # is more than 50% of the total values
        check_missing = (df[self.TESTS] == 0).mean() < 0.5
        # Check if the number of the positive unique values
        # is less than 1% of the total values
        positive_df = df.loc[df[self.TESTS] > 0, self.TESTS]
        try:
            check_unique = (positive_df.nunique() / positive_df.size) >= 0.01
        except ZeroDivisionError:
            return False
        # Result
        return check_zero and check_missing and check_unique

    def _subset_by_area(self, country, province, dataset="COVID-19 Data Hub"):
        """
        Return the subset of "Our World In Data".

        Args:
            country (str): country name
            province (str): province name or "-"
            dataset (str): 'COVID-19 Data Hub' or 'Our World In Data'
        """
        dataset_dict = {
            "COVID-19 Data Hub": self._cleaned_df,
            "Our World In Data": self._cleaned_df_owid,
        }
        df = dataset_dict[dataset].copy()
        return df.loc[(df[self.COUNTRY] == country) & (df[self.PROVINCE] == province)]

    def _subset_select(self, country, province):
        """
        When only "Our World In Data" has sufficient data, the subset of this dataset will be returned.
        If not, "COVID-19 Data Hub" will be selected.

        Args:
            country (str): country name
            province (str): province name or "-"
        """
        # If 'COVID-19 Data Hub' has sufficient data for the area, it will be used
        hub_df = self._subset_by_area(
            country, province, dataset="COVID-19 Data Hub")
        if self._pcr_check_preconditions(hub_df):
            return hub_df
        # If 'Our World In Data' has sufficient data for the area, it will be used
        owid_df = self._subset_by_area(
            country, province, dataset="Our World In Data")
        if self._pcr_check_preconditions(owid_df):
            return owid_df
        # Failed in retrieving sufficient data
        raise PCRIncorrectPreconditionError(
            country=country, province=province, message="Too many missing Tests records")

    def positive_rate(self, country, province=None, window=7, show_figure=True, filename=None):
        """
        Return the PCR rate of a country as a dataframe.

        Args:
            country(str): country name or ISO3 code
            province(str or None): province name
            window (int): window of moving average, >= 1
            show_figure (bool): if True, show the records as a line-plot.
            filename (str): filename of the figure, or None (display figure)

        Raises:
            PCRIncorrectPreconditionError: the dataset has too many missing values

        Returns:
            pandas.DataFrame
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Tests (int): the number of total tests performed
                    - Confirmed (int): the number of confirmed cases
                    - Tests_diff (int): daily tests performed
                    - Confirmed_diff (int): daily confirmed cases
                    - Test_positive_rate (float): positive rate (%) of the daily cases over the total daily tests performed

        Notes:
            If non monotonic records were found for either confirmed cases or tests,
            "with partially complemented tests data" will be added to the title of the figure.
        """
        window = self.ensure_natural_int(window, name="window")
        # Subset with area
        country_alias = self.ensure_country_name(country)
        province = province or self.UNKNOWN
        try:
            subset_df = self._subset_select(country_alias, province)
        except PCRIncorrectPreconditionError:
            raise PCRIncorrectPreconditionError(
                country=country, province=province, message="Too many missing Tests records") from None
        # Process PCR data
        df, is_complemented = self._pcr_processing(subset_df, window)
        # Calculate PCR values
        df[self.PCR_RATE] = df[[self.C_DIFF, self.T_DIFF]].swifter.progress_bar(False).apply(
            lambda x: x[0] / x[1] * 100 if x[1] > self.min_pcr_tests else 0, axis=1)
        if not show_figure:
            return df
        # Create figure
        area = self.area_name(country, province=province)
        comp_status = "\nwith partially complemented tests data" if is_complemented else ""
        line_plot(
            df.set_index(self.DATE)[self.PCR_RATE],
            title=f"{area}: Test positive rate (%) over time {comp_status}",
            ylabel="Test positive rate (%)",
            y_integer=True,
            filename=filename,
            show_legend=False,
        )
        return df

    def subset(self, country, province=None, start_date=None, end_date=None, dataset="COVID-19 Data Hub"):
        """
        Return subset of the country/province and start/end date.

        Args:
            country (str): country name or ISO3 code
            province (str or None): province name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            dataset (str): 'COVID-19 Data Hub' or 'Our World In Data'

        Returns:
            pandas.DataFrame
                Index: reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Tests (int): the number of total tests performed
                    - Confirmed (int): the number of confirmed cases
        """
        country_alias = self.ensure_country_name(country)
        df = self._subset_by_area(
            country=country_alias, province=province, dataset=dataset)
        df = df.drop(
            [self.COUNTRY, self.ISO3, self.PROVINCE], axis=1)
        # Subset with Start/end date
        if start_date is None and end_date is None:
            return df.reset_index(drop=True)
        series = df[self.DATE].copy()
        start_obj = self.date_obj(date_str=start_date, default=series.min())
        end_obj = self.date_obj(date_str=end_date, default=series.max())
        df = df.loc[(start_obj <= series) & (series <= end_obj), :]
        if df.empty:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province,
                start_date=start_date, end_date=end_date)
        return df.reset_index(drop=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from covsirphy.util.error import SubsetNotFoundError
from covsirphy.util.argument import find_args
from covsirphy.cleaning.jhu_data import JHUData
from covsirphy.ode.mbase import ModelBase
from covsirphy.simulation.simulator import ODESimulator


class ExampleData(JHUData):
    """
    Example dataset as a child class of JHUData.

    Args:
        clean_df (pandas.DataFrame or None): cleaned data

            Index:
                - reset index
            Columns:
                - Date (pd.TimeStamp): Observation date
                - Country (pandas.Category): country/region name
                - Province (pandas.Category): province/prefecture/sstate name
                - Confirmed (int): the number of confirmed cases
                - Infected (int): the number of currently infected cases
                - Fatal (int): the number of fatal cases
                - Recovered (int): the number of recovered cases

        tau (int): tau value [min]
        start_date (str): start date, like 22Jan2020
    """

    def __init__(self, clean_df=None, tau=1440, start_date="22Jan2020"):
        if clean_df is None:
            clean_df = pd.DataFrame(columns=self.COLUMNS)
        clean_df = self.ensure_dataframe(
            clean_df, name="clean_df", columns=self.COLUMNS)
        self._raw = clean_df.copy()
        self._cleaned_df = clean_df.copy()
        self._citation = str()
        self.tau = self.ensure_tau(tau)
        self.start_date = self.ensure_date(start_date, name="start_date")
        self._specialized_dict = {}
        self.nondim_dict = {}

    def _model_to_area(self, model=None, country=None, province=None):
        """
        If country is None and model is not None, model name will returned as country name.

        Args:
            model (cs.ModelBase or None): the first ODE model
            country (str or None): country name
            province (str or None): province name

        Raises:
            ValueError: both of country and model are None

        Returns:
            tuple(str, str): country name and province name
        """
        province = province or self.UNKNOWN
        if country is not None:
            return (country, province)
        if model is None:
            raise ValueError("@model or @country must be specified.")
        model = self.ensure_subclass(model, ModelBase, name="model")
        return (model.NAME, province)

    def add(self, model, country=None, province=None, **kwargs):
        """
        Add example data.
        If the country has been registered,
        the start date will be the next data of the registered records.

        Args:
            model (cs.ModelBase): the first ODE model
            country (str or None): country name
            province (str or None): province name
            kwargs: the other keyword arguments of ODESimulator.add()

        Notes:
            If country is None, the name of the model will be used.
            If province is None, '-' will be used.
        """
        # Arguments
        model = self.ensure_subclass(model, ModelBase, name="model")
        arg_dict = model.EXAMPLE.copy()
        arg_dict.update(kwargs)
        population = arg_dict["population"]
        # Area
        country, province = self._model_to_area(
            model=model, country=country, province=province)
        # Start date and y0 values
        df = self._cleaned_df.copy()
        df = df.loc[
            (df[self.COUNTRY] == country) & (df[self.PROVINCE] == province), :
        ]
        if df.empty:
            start_date = self.start_date
        else:
            start_date = df.loc[
                df.index[-1], self.DATE].strftime(self.DATE_FORMAT)
            df = model.tau_free(df, population, tau=None)
            arg_dict[self.Y0_DICT] = {
                k: df.loc[df.index[0], k] for k in model.VARIABLES
            }
        # Simulation
        simulator = ODESimulator(country=country, province=province)
        simulator.add(model=model, **arg_dict)
        # Specialized records
        dim_df = simulator.dim(tau=self.tau, start_date=start_date)
        if country not in self._specialized_dict:
            self._specialized_dict[country] = {}
        self._specialized_dict[country][province] = dim_df.copy()
        # JHU-type records
        restored_df = model.restore(dim_df)
        restored_df[self.COUNTRY] = country
        restored_df[self.PROVINCE] = province
        selected_df = restored_df.loc[:, self.COLUMNS]
        self._cleaned_df = pd.concat(
            [self._cleaned_df, selected_df], axis=0, ignore_index=True
        )
        # Set non-dimensional data
        if country not in self.nondim_dict:
            self.nondim_dict[country] = {}
        nondim_df = simulator.non_dim()
        if province in self.nondim_dict[country]:
            nondim_df_old = self.nondim_dict[country][province].copy()
            nondim_df = pd.concat([nondim_df_old, nondim_df], axis=0)
        self.nondim_dict[country][province] = nondim_df.copy()

    def specialized(self, model=None, country=None, province=None):
        """
        Return dimensional records with model variables.

        Args:
            model (cs.ModelBase or None): the first ODE model
            country (str or None): country name
            province (str or None): province name

        Notes:
            If country is None, the name of the model will be used.
            If province is None, '-' will be used.
        """
        country, province = self._model_to_area(
            model=model, country=country, province=province)
        try:
            return self._specialized_dict[country][province]
        except KeyError:
            raise SubsetNotFoundError(country=country, province=province)

    def non_dim(self, model=None, country=None, province=None):
        """
        Return non-dimensional data.

        Args:
            model (cs.ModelBase or None): the first ODE model
            country (str or None): country name
            province (str or None): province name

        Notes:
            If country is None, the name of the model will be used.
            If province is None, '-' will be used.
        """
        country, province = self._model_to_area(
            model=model, country=country, province=province)
        try:
            return self.nondim_dict[country][province]
        except KeyError:
            raise SubsetNotFoundError(country=country, province=province)

    def subset(self, model=None, country=None, province=None, **kwargs):
        """
        Return the subset of dataset.

        Args:
            model (cs.ModelBase or None): the first ODE model
            country (str or None): country name
            province (str or None): province name
            kwargs: other keyword arguments of JHUData.subset()

        Returns:
            (pandas.DataFrame)
                Index:
                    reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases (> 0)
                    - Susceptible (int): the number of susceptible cases, if calculated

        Notes:
            If country is None, the name of the model will be used.
            If province is None, '-' will be used.
            If @population is not None, the number of susceptible cases will be calculated.
            Records with Recovered > 0 will be selected.
        """
        country, _ = self._model_to_area(
            model=model, country=country, province=province)
        kwargs = find_args([super().subset], **kwargs)
        return super().subset(country=country, province=province, **kwargs)

    def subset_complement(self, **kwargs):
        """
        This is the same as ExampleData.subset().
        Complement will not be done.
        """
        return (self.subset(**kwargs), False)

    def records(self, **kwargs):
        """
        This is the same as ExampleData.subset().
        Complement will not be done.
        """
        return (self.subset(**kwargs), False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import warnings
import covid19dh
import pandas as pd
from covsirphy.util.file import save_dataframe
from covsirphy.cleaning.term import Term
from covsirphy.cleaning.jhu_data import JHUData
from covsirphy.cleaning.oxcgrt import OxCGRTData
from covsirphy.cleaning.population import PopulationData
from covsirphy.cleaning.pcr_data import PCRData


class COVID19DataHub(Term):
    """
    Load datasets retrieved from COVID-19 Data Hub.
    https://covid19datahub.io/

    Args:
        filename (str): CSV filename to save records
    """
    # Citation
    CITATION = '(Secondary source)' \
        ' Guidotti, E., Ardia, D., (2020), "COVID-19 Data Hub",' \
        ' Journal of Open Source Software 5(51):2376, doi: 10.21105/joss.02376.'
    # Class objects of datasets
    OBJ_DICT = {
        "jhu": JHUData,
        "population": PopulationData,
        "oxcgrt": OxCGRTData,
        "pcr": PCRData,
    }

    def __init__(self, filename):
        try:
            self.filepath = Path(filename)
        except TypeError:
            raise TypeError(
                f"@filename should be a path-like object, but {filename} was applied.")
        self.filepath.parent.mkdir(exist_ok=True, parents=True)
        self.primary_list = None

    def load(self, name="jhu", force=True, verbose=1):
        """
        Load the datasets of COVID-19 Data Hub.

        Args:
            name (str): name of dataset, "jhu", "population", "oxcgrt" or "pcr"
            force (bool): if True, always download the dataset from the server
            verbose (int): level of verbosity

        Returns:
            covsirphy.CleaningBase: the dataset

        Notes:
            If @verbose is 2, detailed citation list will be shown when downloading.
            If @verbose is 1, how to show the list will be explained.
            Citation of COVID-19 Data Hub will be set as JHUData.citation etc.
        """
        if force and self.filepath.exists():
            self.filepath.unlink()
        if not self.filepath.exists():
            raw_df = self._retrieve(verbose=verbose)
            save_dataframe(raw_df, self.filepath, index=False)
        if name not in self.OBJ_DICT:
            raise KeyError(
                f"@name must be {', '.join(list(self.OBJ_DICT.keys()))}, but {name} was applied.")
        return self.OBJ_DICT[name](self.filepath, citation=self.CITATION)

    def _retrieve(self, verbose=1):
        """
        Retrieve datasets from COVID-19 Data Hub.
        Level 1 (country) and level2 (province) will be used and combined to a dataframe.

        Args:
            verbose (int): level of verbosity

        Notes:
            If @verbose is 2, detailed citation list will be shown when downloading.
            If @verbose is 1, how to show the list will be explained.
        """
        # Country level
        if verbose:
            print(
                "Retrieving datasets from COVID-19 Data Hub: https://covid19datahub.io/")
        c_df, p_df, self.primary_list = self._download()
        # Change column names and select columns to use
        # All columns: https://covid19datahub.io/articles/doc/data.html
        col_dict = {
            "date": "ObservationDate",
            "tests": self.TESTS,
            "confirmed": self.C,
            "recovered": self.R,
            "deaths": "Deaths",
            "population": self.N,
            "iso_alpha_3": self.ISO3,
            "administrative_area_level_2": "Province/State",
            "administrative_area_level_1": "Country/Region",
        }
        columns = list(col_dict.values()) + OxCGRTData.OXCGRT_VARIABLES_RAW
        # Merge the datasets
        c_df = c_df.rename(col_dict, axis=1).loc[:, columns]
        p_df = p_df.rename(col_dict, axis=1).loc[:, columns]
        df = pd.concat([c_df, p_df], axis=0, ignore_index=True)
        if verbose:
            if isinstance(verbose, int) and verbose >= 2:
                print("\nDetailed citaition list:")
                print(self.primary_list)
            else:
                print("\nPlease set verbose=2 to see the detailed citation list.")
            print("\n")
        return df

    def _download(self):
        """
        Retrieve dataset and citation list from COVID-19 Data Hub.
        Citation list will be saved to self.

        Returns:
            tuple:
                pandas.DataFrame: dataset at country level
                pandas.DataFrame: dataset at province level
                str: the list of primary sources

        Notes:
            For some countries, province-level data is included.
        """
        warnings.simplefilter("ignore", ResourceWarning)
        c_res = covid19dh.covid19(
            country=None, level=1, verbose=False, raw=False)
        p_res = covid19dh.covid19(
            country=None, level=2, verbose=False, raw=False)
        try:
            c_df, c_cite = c_res
            p_df, p_cite = p_res
        except ValueError:
            # covid19dh <= 1.14
            c_df, c_cite = c_res.copy(), covid19dh.cite(c_res)
            p_df, p_cite = p_res.copy(), covid19dh.cite(p_res)
            citations = list(dict.fromkeys(c_cite + p_cite))
            return (c_df, p_df, "\n".join(citations))
        # Citation
        cite = pd.concat([c_cite, p_cite], axis=0, ignore_index=True)
        cite = cite.loc[:, ["title", "year", "url"]]
        cite = cite.sort_values(["year", "url"], ascending=[False, True])
        cite.drop_duplicates(subset="title", inplace=True)
        series = cite.apply(lambda x: f"{x[0]} ({x[1]}), {x[2]}", axis=1)
        return (c_df, p_df, "\n".join(series.tolist()))

    @property
    def primary(self):
        """
        str: the list of primary sources.
        """
        return self.primary_list or self._download()[-1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import pandas as pd


def save_dataframe(df, filename, index=True):
    """
    Save dataframe as a CSV file.

    Args:
        df (pd.DataFrame): the dataframe
        filename (str or None): CSV filename
        index (bool): if True, include index column.

    Notes:
        If @filename is None or OSError was raised, the dataframe will not be saved.
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError(
            f"@df should be a pandas.DataFrame, but {type(df)} was applied.")
    try:
        filepath = Path(filename)
        df.to_csv(filepath, index=index)
    except (TypeError, OSError):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from covsirphy import PhaseUnit
from covsirphy import Term, SIR, Estimator, UnExecutedError


class TestPhaseUnit(object):
    def test_start(self):
        unit = PhaseUnit("01Jan2020", "01Feb2020", 1000)
        assert str(unit) == "Phase (01Jan2020 - 01Feb2020)"
        summary_dict = unit.to_dict()
        assert summary_dict[Term.START] == "01Jan2020"
        assert summary_dict[Term.END] == "01Feb2020"
        assert summary_dict[Term.N] == 1000
        summary_df = unit.summary()
        assert set(summary_df.columns) == set([Term.START, Term.END, Term.N])
        with pytest.raises(NotImplementedError):
            assert unit == "phase"
        unit2 = PhaseUnit("01Jan2020", "01Feb2020", 100000)
        assert unit == unit2
        unit3 = PhaseUnit("01Jan2020", "01Mar2020", 1000)
        assert unit != unit3
        assert "31Dec2019" not in unit
        assert "01Jan2020" in unit
        assert "15Jan2020" in unit
        assert "01Feb2020" in unit
        assert "01Mar2020" not in unit

    def test_sort(self):
        unit1 = PhaseUnit("01Jan2020", "01Feb2020", 1000)
        unit2 = PhaseUnit("02Feb2020", "01Mar2020", 1000)
        unit3 = PhaseUnit("02Mar2020", "01Apr2020", 1000)
        unit4 = PhaseUnit("02Mar2020", "01Apr2020", 1000)
        unit5 = PhaseUnit("01Jan2020", "01Apr2020", 1000)
        assert unit3 == unit4
        assert unit1 != unit2
        assert unit1 < unit2
        assert unit3 > unit1
        assert not unit3 > unit4
        assert unit3 <= unit4
        assert unit1 < "02Feb2020"
        assert unit1 <= "01Feb2020"
        assert unit1 > "31Dec2019"
        assert unit1 >= "01Jan2020"
        assert sorted([unit3, unit1, unit2]) == [unit1, unit2, unit3]
        assert str(unit1 + unit2) == "Phase (01Jan2020 - 01Mar2020)"
        assert str(unit5 - unit1) == "Phase (02Feb2020 - 01Apr2020)"
        assert str(unit5 - unit4) == "Phase (01Jan2020 - 01Mar2020)"
        assert set([unit1, unit3, unit4]) == set([unit1, unit3])

    def test_enable(self):
        unit = PhaseUnit("01Jan2020", "01Feb2020", 1000)
        assert bool(unit)
        unit.disable()
        assert not bool(unit)
        unit.enable()
        assert bool(unit)

    def test_definition_property(self):
        unit = PhaseUnit("01Jan2020", "01Feb2020", 1000)
        assert unit.start_date == "01Jan2020"
        with pytest.raises(AttributeError):
            unit.start_date = "10Jan2020"
        assert unit.end_date == "01Feb2020"
        with pytest.raises(AttributeError):
            unit.end_date = "10Jan2020"
        assert unit.population == 1000
        with pytest.raises(AttributeError):
            unit.population = 100000

    def test_tau(self):
        unit = PhaseUnit("01Jan2020", "01Feb2020", 1000)
        assert unit.tau is None
        with pytest.raises(ValueError):
            unit.tau = 1000
        unit.tau = 1440
        assert unit.tau == 1440
        with pytest.raises(AttributeError):
            unit.tau = 720

    def test_set_ode(self):
        unit = PhaseUnit("01Jan2020", "01Feb2020", 1000)
        unit.set_ode(model=SIR, tau=720, rho=0.2)
        assert issubclass(unit.model, SIR)
        summary_dict = unit.to_dict()
        assert summary_dict[Term.TAU] == 720
        assert summary_dict[Term.ODE] == SIR.NAME
        assert summary_dict["rho"] == 0.2
        assert summary_dict["sigma"] is None

    def test_set_tau(self):
        unit = PhaseUnit("01Jan2020", "01Feb2020", 1000)
        unit.set_ode(model=None, tau=240, rho=0.006)
        assert unit.tau == 240
        assert "rho" not in unit.to_dict()

    @pytest.mark.parametrize("country", ["Japan"])
    def test_estimate(self, jhu_data, population_data, country):
        # Dataset
        population = population_data.value(country)
        record_df = jhu_data.subset(country, population=population)
        # Parameter estimation
        unit = PhaseUnit("27May2020", "27Jun2020", population)
        with pytest.raises(NameError):
            unit.estimate()
        unit.set_ode(model=SIR)
        with pytest.raises(ValueError):
            unit.estimate()
        unit.record_df = record_df
        assert set(unit.record_df.columns) == set(Term.SUB_COLUMNS)
        unit.estimate(timeout=20)
        # Check results
        assert isinstance(unit.estimator, Estimator)
        assert set(SIR.PARAMETERS).issubset(unit.to_dict())
        assert set(SIR.DAY_PARAMETERS).issubset(unit.to_dict())
        cols = [Term.ODE, Term.RMSLE, Term.TRIALS, Term.RUNTIME]
        assert set(cols).issubset(unit.to_dict())
        assert None not in unit.to_dict().values()

    @pytest.mark.parametrize("country", ["Japan"])
    def test_estimate_with_fixed(self, jhu_data, population_data, country):
        # Dataset
        population = population_data.value(country)
        record_df = jhu_data.subset(country, population=population)
        # Setting
        unit = PhaseUnit("27May2020", "27Jun2020", population)
        unit.set_ode(model=SIR, tau=360, rho=0.01)
        # Parameter estimation
        unit.estimate(record_df=record_df)
        assert unit.tau == unit.to_dict()[Term.TAU] == 360
        assert unit.to_dict()["rho"] == 0.01

    @pytest.mark.parametrize("country", ["Japan"])
    def test_simulate(self, jhu_data, population_data, country):
        # Dataset
        population = population_data.value(country)
        record_df = jhu_data.subset(country, population=population)
        # Parameter setting
        unit = PhaseUnit("27May2020", "27Jun2020", population)
        with pytest.raises(UnExecutedError):
            unit3 = PhaseUnit("27May2020", "27Jun2020", population)
            unit3.set_ode(model=SIR, tau=240, rho=0.006)
            unit3.simulate()
        unit.set_ode(model=SIR, tau=240, rho=0.006, sigma=0.011)
        summary_dict = unit.to_dict()
        assert summary_dict[Term.RT] == 0.55
        assert summary_dict["1/beta [day]"] == 27
        assert summary_dict["1/gamma [day]"] == 15
        # Set initial values
        unit.set_y0(record_df)
        # Simulation
        sim_df = unit.simulate()
        assert set(sim_df.columns) == set(Term.NLOC_COLUMNS)

    @pytest.mark.parametrize("country", ["Japan"])
    def test_simulate_y0_direct(self, jhu_data, population_data, country):
        # Dataset
        population = population_data.value(country)
        # Parameter setting
        unit = PhaseUnit("27May2020", "27Jun2020", population)
        unit.set_ode(model=SIR, tau=240, rho=0.006, sigma=0.011)
        # Simulation
        y0_dict = {
            "Susceptible": 126512455, "Infected": 1806, "Fatal or Recovered": 14839
        }
        sim_df = unit.simulate(y0_dict=y0_dict)
        assert set(sim_df.columns) == set(Term.NLOC_COLUMNS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import functools
import warnings
import numpy as np
from datetime import timedelta
from covsirphy.cleaning.term import Term


class JHUDataComplementHandler(Term):
    """
    Complement JHU dataset, if necessary.

    Args:
        recovery_period (int): expected value of recovery period [days]
        interval (int): expected update interval of the number of recovered cases [days]
        max_ignored (int): Max number of recovered cases to be ignored [cases]
        max_ending_unupdated (int) : Max number of days to apply full complement,
                         where max recovered cases are not updated [days]

    Notes:
        To add new complement solutions, we need to update cls.STATUS_NAME_DICT and self._protocol().
        Status names with high socres will be prioritized when status code will be determined.
        Status code: 'fully complemented recovered data' and so on as noted in self.run() docstring.
    """
    RAW_COLS = [Term.C, Term.F, Term.R]
    # Kind of complement: {score: name}
    STATUS_NAME_DICT = {
        1: "sorting",
        2: "monotonic increasing",
        3: "partially",
        4: "fully",
    }

    def __init__(self, recovery_period, interval, max_ignored, max_ending_unupdated):
        # Arguments for complement
        self.recovery_period = self.ensure_natural_int(
            recovery_period, name="recovery_period")
        self.interval = self.ensure_natural_int(interval, name="interval")
        self.max_ignored = self.ensure_natural_int(
            max_ignored, name="max_ignored")
        self.max_ending_unupdated = self.ensure_natural_int(
            max_ending_unupdated, name="max_ending_unupdated")

    def _protocol(self):
        """
        Return the list of complement solutions and scores.

        Returns:
            list[str/None, function, int]: nested list of variables to be updated, methods and scores
        """
        return [
            (None, self._pre_processing, 0),
            (self.C, functools.partial(self._monotonic, variable=self.C), 2),
            (self.F, functools.partial(self._monotonic, variable=self.F), 2),
            (self.R, functools.partial(self._monotonic, variable=self.R), 2),
            (self.R, self._recovered_full, 4),
            (self.R, self._recovered_sort, 1),
            (self.R, functools.partial(self._monotonic, variable=self.R), 2),
            (self.R, self._recovered_partial_ending, 3),
            (self.R, self._recovered_partial, 3),
            (self.R, functools.partial(self._monotonic, variable=self.R), 2),
            (self.R, self._recovered_sort, 1),
            (None, self._post_processing, 0)
        ]

    def run(self, subset_df):
        """
        Perform complement.

        Args:
            subset_df (pandas.DataFrame): Subset of records
                Index: reset index
                Columns:
                    - Date (pd.TimeStamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - The other columns will be ignored

        Returns:
            tuple(pandas.DataFrame, str):
                pandas.DataFrame:
                    Index: reset index
                    Columns:
                        - Date(pd.TimeStamp): Observation date
                        - Confirmed(int): the number of confirmed cases
                        - Infected(int): the number of currently infected cases
                        - Fatal(int): the number of fatal cases
                        - Recovered (int): the number of recovered cases
                str: status code

        Notes:
            Status code will be selected from:
                - '' (not complemented)
                - 'monotonic increasing complemented confirmed data'
                - 'monotonic increasing complemented fatal data'
                - 'monotonic increasing complemented recovered data'
                - 'fully complemented recovered data'
                - 'partially complemented recovered data'
        """
        self.ensure_dataframe(
            subset_df, name="subset_df", columns=[self.DATE, *self.RAW_COLS])
        # Initialize
        after_df = subset_df.copy()
        status_dict = dict.fromkeys(self.RAW_COLS, 0)
        # Perform complement one by one
        for (variable, func, score) in self._protocol():
            before_df, after_df = after_df.copy(), func(after_df)
            if after_df.equals(before_df) or variable is None:
                continue
            status_dict[variable] = max(status_dict[variable], score)
        # Create status code
        status_list = [
            f"{self.STATUS_NAME_DICT[score]} complemented {v.lower()} data"
            for (v, score) in status_dict.items() if score]
        status = " and \n".join(status_list)
        return (after_df, status)

    def _pre_processing(self, subset_df):
        """
        Select Confirmed/Fatal/Recovered class from the dataset.

        Args:
            subset_df (pandas.DataFrame): Subset of records
                Index: reset index
                Columns: Date, Confirmed, Fatal, Recovered (the others will be ignored)

        Returns:
            pandas.DataFrame
                Index: Date (pandas.TimeStamp)
                Columns: Confirmed, Fatal, Recovered
        """
        return subset_df.set_index(self.DATE).loc[:, self.RAW_COLS]

    def _post_processing(self, df):
        """
        Select Confirmed/Fatal/Recovered class from the dataset.

        Args:
            df (pandas.DataFrame)
                Index: Date (pandas.TimeStamp)
                Columns: Confirmed, Fatal, Recovered

        Returns:
            pandas.DataFrame: complemented records
                Index: reset index
                Columns: Date (pandas.TimeStamp), Confirmed, Infected, Fatal, Recovered (int)
        """
        df = df.astype(np.int64)
        df[self.CI] = df[self.C] - df[self.F] - df[self.R]
        df = df.loc[:, self.VALUE_COLUMNS]
        return df.reset_index()

    def _monotonic(self, df, variable):
        """
        Force the variable show monotonic increasing.

        Args:
            df (pandas.DataFrame):
                Index: Date (pandas.TimeStamp)
                Columns: Confirmed, Fatal, Recovered
            variable (str): variable name to show monotonic increasing

        Returns:
            pandas.DataFrame: complemented records
                Index: Date (pandas.TimeStamp)
                Columns: Confirmed, Fatal, Recovered
        """
        warnings.simplefilter("ignore", UserWarning)
        # Whether complement is necessary or not
        if df[variable].is_monotonic_increasing:
            return df
        # Complement
        decreased_dates = df[df[variable].diff() < 0].index.tolist()
        for date in decreased_dates:
            # Raw value on the decreased date
            raw_last = df.loc[date, variable]
            # Extrapolated value on the date
            series = df.loc[:date, variable]
            series.iloc[-1] = None
            series.interpolate(method="spline", order=1, inplace=True)
            series.fillna(method="ffill", inplace=True)
            # Reduce values to the previous date
            df.loc[:date, variable] = series * raw_last / series.iloc[-1]
            df[variable] = df[variable].fillna(0).astype(np.int64)
        return df

    def _recovered_full(self, df):
        """
        Estimate the number of recovered cases with the value of recovery period.

        Args:
            df (pandas.DataFrame):
                Index: Date (pandas.TimeStamp)
                Columns: Confirmed, Fatal, Recovered

        Returns:
            pandas.DataFrame: complemented records
                Index: Date (pandas.TimeStamp)
                Columns: Confirmed, Fatal, Recovered
        """
        # Whether complement is necessary or not
        if df[self.R].max() > self.max_ignored:
            # Necessary if sum of recovered is more than 99%
            # or less than 1% of sum of recovered minus infected when outbreaking
            sel_C1 = df[self.C] > self.max_ignored
            sel_R1 = df[self.R] > self.max_ignored
            sel_2 = df[self.C].diff().diff().rolling(14).mean() > 0
            cf_diff = (df[self.C] - df[self.F]).rolling(14).mean()
            sel_3 = df[self.R] < 0.01 * cf_diff
            sel_4 = df[self.R] > 0.99 * cf_diff
            s_df_1 = df.loc[sel_C1 & sel_2 & sel_4]
            s_df_2 = df.loc[sel_R1 & sel_2 & sel_3]
            if s_df_1.empty and s_df_2.empty:
                return df
        # Estimate recovered records
        df[self.R] = (df[self.C] - df[self.F]).shift(
            periods=self.recovery_period, freq="D")
        return df

    def _recovered_partial_ending(self, df):
        """
        If ending recovered values do not change for more than applied 'self.interval' days
        after reached 'self.max_ignored' cases, apply full complement only to these
        unupdated values and keep the previous valid ones.
        _recovered_partial() does not handle well such values, because
        they are not in-between values and interpolation generates only similar values,
        but small compared to confirmed cases.

        Args:
            df (pandas.DataFrame):
                Index: Date (pandas.TimeStamp)
                Columns: Confirmed, Fatal, Recovered

        Returns:
            pandas.DataFrame: complemented records
                Index: Date (pandas.TimeStamp)
                Columns: Confirmed, Fatal, Recovered
        """
        # Whether complement is necessary or not
        r_max = df[self.R].max()
        sel_0 = (df[self.R] == r_max).sum() > self.max_ending_unupdated
        if not (r_max and sel_0):
            # full complement will be handled in _recovered_full()
            return df
        # Complement any ending unupdated values.
        # Fully complement last records that are not updated
        # for more than max_ending_unupdated days;
        # keep previous valid values as they are
        min_index = df[self.R].idxmax() + timedelta(days=1)
        df.loc[min_index:, self.R] = (df[self.C] - df[self.F]).shift(
            periods=self.recovery_period, freq="D").loc[min_index:]
        return df

    def _recovered_partial(self, df):
        """
        If recovered values do not change for more than applied 'self.interval' days
        after reached 'self.max_ignored' cases, interpolate the recovered values.

        Args:
            df (pandas.DataFrame):
                Index: Date (pandas.TimeStamp)
                Columns: Confirmed, Fatal, Recovered

        Returns:
            pandas.DataFrame: complemented records
                Index: Date (pandas.TimeStamp)
                Columns: Confirmed, Fatal, Recovered
        """
        # Whether complement is necessary or not
        series = df.loc[df[self.R] > self.max_ignored, self.R]
        max_frequency = series.value_counts().max()
        if max_frequency <= self.interval:
            return df
        # Complement in-between values
        df.loc[df.duplicated([self.R], keep="last"), self.R] = None
        df[self.R].interpolate(
            method="linear", inplace=True, limit_direction="both")
        df[self.R] = df[self.R].fillna(method="bfill")
        return df

    def _recovered_sort(self, df):
        """
        Sort the absolute values of recovered data.

        Args:
            df (pandas.DataFrame):
                Index: Date (pandas.TimeStamp)
                Columns: Confirmed, Fatal, Recovered

        Returns:
            pandas.DataFrame: complemented records
                Index: Date (pandas.TimeStamp)
                Columns: Confirmed, Fatal, Recovered
        """
        df.loc[:, self.R] = sorted(df[self.R].abs())
        df[self.R].interpolate(method="time", inplace=True)
        df[self.R] = df[self.R].fillna(0).round()
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import warnings
import country_converter as coco
from dask import dataframe as dd
import pandas as pd
from covsirphy.util.error import deprecate, SubsetNotFoundError
from covsirphy.cleaning.term import Term


class CleaningBase(Term):
    """
    Basic class for data cleaning.

    Args:
        filename (str or None): CSV filename of the dataset
        citation (str or None): citation

    Returns:
        If @filename is None, empty dataframe will be set as raw data.
        If @citation is None, citation will be empty string.
    """

    def __init__(self, filename, citation=None):
        warnings.simplefilter("ignore", DeprecationWarning)
        if filename is None:
            self._raw = pd.DataFrame()
            self._cleaned_df = pd.DataFrame()
        else:
            Path(filename).parent.mkdir(exist_ok=True, parents=True)
            self._raw = dd.read_csv(
                filename, dtype={"Province/State": "object"}
            ).compute()
            self._cleaned_df = self._cleaning()
        self._citation = citation or ""

    @property
    def raw(self):
        """
        pandas.DataFrame: raw data
        """
        return self._raw

    @raw.setter
    def raw(self, dataframe):
        """
        pandas.DataFrame: raw dataset
        """
        self._raw = self.ensure_dataframe(dataframe, name="dataframe")

    @staticmethod
    def load(urlpath, header=0, columns=None, dtype="object"):
        """
        Load a local/remote file.

        Args:
            urlpath (str or pathlib.Path): filename or URL
            header (int): row number of the header
            columns (list[str]): columns to use
            dtype (str or dict[str]): data type for the dataframe or specified columns

        Returns:
            pd.DataFrame: raw dataset
        """
        kwargs = {
            "low_memory": False, "dtype": dtype, "header": header, "usecols": columns}
        try:
            return dd.read_csv(urlpath, blocksize=None, **kwargs).compute()
        except (FileNotFoundError, UnicodeDecodeError, pd.errors.ParserError):
            return pd.read_csv(urlpath, **kwargs)

    def cleaned(self):
        """
        Return the cleaned dataset.

        Notes:
            Cleaning method is defined by CleaningBase._cleaning() method.

        Returns:
            pandas.DataFrame: cleaned data
        """
        return self._cleaned_df

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Returns:
            pandas.DataFrame: cleaned data
        """
        return self._raw.copy()

    @property
    def citation(self):
        """
        str: citation/description of the dataset
        """
        return self._citation

    @citation.setter
    def citation(self, description):
        self._citation = str(description)

    def ensure_country_name(self, country):
        """
        Ensure that the country name is correct.
        If not, the correct country name will be found.

        Args:
            country (str): country name

        Returns:
            str: country name
        """
        df = self.ensure_dataframe(
            self._cleaned_df, name="the cleaned dataset", columns=[self.COUNTRY])
        selectable_set = set(df[self.COUNTRY].unique())
        # return country name as-is if selectable
        if country in selectable_set:
            return country
        # Convert country name
        converted = coco.convert(country, to="name_short", not_found=None)
        # Additional abbr
        abbr_dict = {
            "Congo Republic": "Republic of the Congo",
            "DR Congo": "Democratic Republic of the Congo",
            "UK": "United Kingdom",
            "Vatican": "Holy See",
        }
        name = abbr_dict.get(converted, converted)
        # Return the name if registered in the dataset
        if name in selectable_set:
            return name
        raise SubsetNotFoundError(country=country, country_alias=name)

    @deprecate("CleaningBase.iso3_to_country()", new="CleaningBase.ensure_country_name()")
    def iso3_to_country(self, iso3_code):
        """
        Convert ISO3 code to country name if records are available.

        Args:
            iso3_code (str): ISO3 code or country name

        Returns:
            str: country name

        Notes:
            If ISO3 codes are not registered, return the string as-si @iso3_code.
        """
        return self.ensure_country_name(iso3_code)

    def country_to_iso3(self, country):
        """
        Convert country name to ISO3 code if records are available.

        Args:
            country (str): country name

        Raises:
            KeyError: ISO3 code of the country is not registered

        Returns:
            str: ISO3 code or "---" (when unknown)
        """
        name = self.ensure_country_name(country)
        return coco.convert(name, to="ISO3", not_found="---")

    @classmethod
    def area_name(cls, country, province=None):
        """
        Return area name of the country/province.

        Args:
            country (str): country name or ISO3 code
            province (str): province name

        Returns:
            str: area name

        Notes:
            If province is None or '-', return country name.
            If not, return the area name, like 'Japan/Tokyo'
        """
        if province in [None, cls.UNKNOWN]:
            return country
        return f"{country}{cls.SEP}{province}"

    def _subset_by_area(self, country, province=None):
        """
        Return subset for the country/province.

        Args:
            country (str): country name
            province (str or None): province name

        Returns:
            pandas.DataFrame: subset for the country/province
        """
        # Country level
        country = self.ensure_country_name(country)
        df = self.ensure_dataframe(
            self._cleaned_df, name="the cleaned dataset", columns=[self.COUNTRY])
        df = df.loc[df[self.COUNTRY] == country, :]
        # Province level
        province = province or self.UNKNOWN
        if self.PROVINCE not in df.columns and province == self.UNKNOWN:
            return df
        df = self.ensure_dataframe(
            df, "the cleaned dataset", columns=[self.PROVINCE])
        return df.loc[df[self.PROVINCE] == province, :]

    def subset(self, country, province=None, start_date=None, end_date=None):
        """
        Return subset of the country/province and start/end date.

        Args:
            country (str): country name or ISO3 code
            province (str or None): province name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020

        Returns:
            pandas.DataFrame
                Index: reset index
                Columns: without ISO3, Country, Province column
        """
        country_alias = self.ensure_country_name(country)
        df = self._subset_by_area(country=country, province=province)
        df = df.drop(
            [self.COUNTRY, self.ISO3, self.PROVINCE], axis=1, errors="ignore")
        if df.empty:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province)
        # Subset with Start/end date
        if start_date is None and end_date is None:
            return df.reset_index(drop=True)
        df = self.ensure_dataframe(
            df, name="the cleaned dataset", columns=[self.DATE])
        series = df[self.DATE].copy()
        start_obj = self.date_obj(date_str=start_date, default=series.min())
        end_obj = self.date_obj(date_str=end_date, default=series.max())
        df = df.loc[(start_obj <= series) & (series <= end_obj), :]
        if df.empty:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province,
                start_date=start_date, end_date=end_date)
        return df.reset_index(drop=True)

    def subset_complement(self, country, **kwargs):
        """
        Return the subset. If necessary, complemention will be performed.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def records(self, country, province=None, start_date=None, end_date=None,
                auto_complement=True, **kwargs):
        """
        Return the subset. If necessary, complemention will be performed.

        Args:
            country (str): country name or ISO3 code
            province (str or None): province name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            auto_complement (bool): if True and necessary, the number of cases will be complemented
            kwargs: the other arguments of complement

        Returns:
            pandas.DataFrame
                Index: reset index
                Columns:
                    without ISO3, Country, Province column
        """
        country_alias = self.ensure_country_name(country)
        subset_arg_dict = {
            "country": country, "province": province, "start_date": start_date, "end_date": end_date}
        if auto_complement:
            try:
                df, is_complemented = self.subset_complement(
                    **subset_arg_dict, **kwargs)
                if not df.empty:
                    return (df, is_complemented)
            except NotImplementedError:
                pass
        try:
            return (self.subset(**subset_arg_dict), False)
        except SubsetNotFoundError:
            raise SubsetNotFoundError(
                country=country, country_alias=country_alias, province=province,
                start_date=start_date, end_date=end_date) from None

    def countries(self):
        """
        Return names of countries where records are registered.

        Raises:
            KeyError: Country names are not registered in this dataset

        Returns:
            list[str]: list of country names
        """
        df = self.ensure_dataframe(
            self._cleaned_df, name="the cleaned dataset", columns=[self.COUNTRY])
        return list(df[self.COUNTRY].unique())

    def total(self):
        """
        Calculate total values of the cleaned dataset.
        """
        raise NotImplementedError
